# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from azure.cosmos.aio import CosmosClient
import azure.cosmos.exceptions as exceptions
from azure.cosmos.http_constants import StatusCodes
from azure.cosmos.partition_key import PartitionKey
import datetime

import asyncio
import config

# ----------------------------------------------------------------------------------------------------------
# Prerequisites -
#
# 1. An Azure Cosmos account -
#    https:#azure.microsoft.com/documentation/articles/documentdb-create-account/
#
# 2. Microsoft Azure Cosmos PyPi package -
#    https://pypi.python.org/pypi/azure-cosmos/
# ----------------------------------------------------------------------------------------------------------
# Sample - demonstrates the basic CRUD operations on a Item resource for Azure Cosmos
# ----------------------------------------------------------------------------------------------------------

HOST = config.settings['host']
MASTER_KEY = config.settings['master_key']
DATABASE_ID = config.settings['database_id']
CONTAINER_ID = config.settings['container_id']
CONTAINER_MH_ID = config.settings['container_mh_id']


async def create_items(container):
    print('Creating Items')
    print('\n1.1 Create Item\n')

    # Create a SalesOrder object. This object has nested properties and various types including numbers, DateTimes and strings.
    # This can be saved as JSON as is without converting into rows/columns.
    sales_order = get_sales_order("SalesOrder1")
    await container.create_item(body=sales_order)

    # As your app evolves, let's say your object has a new schema. You can insert SalesOrderV2 objects without any
    # changes to the database tier.
    sales_order2 = get_sales_order_v2("SalesOrder2")
    await container.create_item(body=sales_order2)


async def read_item(container, doc_id):
    print('\n1.2 Reading Item by Id\n')

    # Note that Reads require a partition key to be specified.
    response = await container.read_item(item=doc_id, partition_key=doc_id)

    print('Item read by Id {0}'.format(doc_id))
    print('Account Number: {0}'.format(response.get('account_number')))
    print('Subtotal: {0}'.format(response.get('subtotal')))


async def read_items(container):
    print('\n1.3 - Reading all items in a container\n')

    # NOTE: Use MaxItemCount on Options to control how many items come back per trip to the server
    #       Important to handle throttles whenever you are doing operations such as this that might
    #       result in a 429 (throttled request)
    read_all_items_response = container.read_all_items(max_item_count=10)

    # The asynchronous client returns an asynchronous iterator object for methods that
    # return several items, so attempting to cast this object into a list directly will
    # throw an error; instead, iterate over the items using an async for loop like shown
    # here and in the query_items() method below. We also do not await read_all() because
    # it doesn't deal with partition key logic the way query_items() does
    item_list = [item async for item in read_all_items_response]

    print('Found {0} items'.format(item_list.__len__()))

    for doc in item_list:
        print('Item Id: {0}'.format(doc.get('id')))

    # Alternatively, you can directly iterate over the asynchronous iterator without building a separate
    # list if you don't need the ordering or indexing capabilities
    async for item in read_all_items_response:
        print(item.get('id'))


async def query_items(container, doc_id):
    print('\n1.4 Querying for an  Item by Id\n')

    # The enable_cross_partition_query option is not used by the async client - this is the default behavior
    # In this case, we do have to await the asynchronous iterator object since logic
    # within the query_items() method makes network calls to verify the partition key
    # definition in the container
    query_items_response = container.query_items(
        query="SELECT * FROM r WHERE r.id=@id",
        parameters=[
            {"name": "@id", "value": doc_id}
        ]
    )

    items = [item async for item in query_items_response]

    print('Item queried by Id {0}'.format(items[0].get("id")))


async def query_items_with_continuation_token(container):
    print('\n1.5 Querying for Items using Pagination and Continuation Token\n')

    # When using Pagination, max_item_count will limit the number of items in each page
    query_iterable = container.query_items(
        query="SELECT * FROM r",
        max_item_count=1
    )

    item_pages = query_iterable.by_page()
    first_page = await anext(item_pages)  # type: ignore[name-defined]  # cspell:disable-line
    continuation_token = item_pages.continuation_token

    # Other code logic where you only need the first page of results would go here

    # Now we use the continuation token from the first page to pick up where we left off and
    # access the second page of items
    items_from_continuation = query_iterable.by_page(continuation_token)
    second_page_items_with_continuation = \
        [i async for i in await anext(items_from_continuation)]  # type: ignore[name-defined]  # cspell:disable-line

    print('The single items in the second page are {}'.format(second_page_items_with_continuation[0].get("id")))


async def query_items_single_partition_with_pagination(container):
    print('\n1.5a Querying with Pagination - Demonstrating max_item_count and Counting Results\n')
    
    # max_item_count controls how many items are returned per page, not the total number of results
    # This is useful for controlling memory usage and processing items in batches
    max_items_per_page = 5

    # Specify a partition key to query within a single partition
    partition_key_value = "SalesOrder1"  # Or any partition key value you want
    
    query_iterable = container.query_items(
        query="SELECT * FROM c",
        partition_key=partition_key_value,  # Query single partition
        max_item_count=max_items_per_page
    )
    
    # Iterate through pages and count both pages and total items
    total_item_count = 0
    page_count = 0
    
    item_pages = query_iterable.by_page()
    async for page in item_pages:
        page_count += 1
        items_in_page = [item async for item in page]
        items_in_current_page = len(items_in_page)
        total_item_count += items_in_current_page
        
        print(f'Page {page_count}: Retrieved {items_in_current_page} items (max per page: {max_items_per_page})')
        
        # Process items in this page
        for item in items_in_page:
            # Do something with each item
            pass
    
    print(f'\nTotal pages processed: {page_count}')
    print(f'Total items retrieved: {total_item_count}')
    print(f'Note: max_item_count limits items PER PAGE, not total results\n')


async def query_items_cross_partition_with_pagination(container):
    print('\n1.5b Cross-Partition Query with Pagination\n')
    
    # When querying across partitions, max_item_count still controls page size
    # but the results are gathered from multiple partitions
    max_items_per_page = 3
    
    query_iterable = container.query_items(
        query="SELECT * FROM c ORDER BY c._ts",  # Order by timestamp across all partitions
        max_item_count=max_items_per_page
    )
    
    total_item_count = 0
    page_count = 0
    
    item_pages = query_iterable.by_page()
    async for page in item_pages:
        page_count += 1
        items_in_page = [item async for item in page]
        total_item_count += len(items_in_page)
        
        print(f'Page {page_count}: {len(items_in_page)} items from across partitions')
    
    print(f'\nCross-partition query completed:')
    print(f'  - Pages: {page_count}')
    print(f'  - Total items: {total_item_count}')

async def query_items_with_feed_ranges_and_pagination(container):
    print('\n1.5c Querying with Feed Range and Pagination\n')
    # We again use max_item_count to control page size
    max_items_per_page = 3

    # First we fetch the relevant feed ranges for the container - feed ranges represent logical partitions
    # or ranges of partition key values, and can be used to query as well
    feed_ranges = container.read_feed_ranges()

    # For this example, we will just use the first feed range
    feed_ranges_list = [feed_range async for feed_range in feed_ranges]
    query_iterable = container.query_items(
        query="SELECT * FROM c",
        feed_range=feed_ranges_list[0],  # Query specific feed range
        max_item_count=max_items_per_page
    )

    # Iterate through pages and count both pages and total items
    total_item_count = 0
    page_count = 0

    item_pages = query_iterable.by_page()
    async for page in item_pages:
        page_count += 1
        items_in_page = [item async for item in page]
        total_item_count += len(items_in_page)

        print(f'Page {page_count}: {len(items_in_page)} items from across partitions')

    print(f'\nCross-partition query completed:')
    print(f'  - Pages: {page_count}')
    print(f'  - Total items: {total_item_count}')

async def replace_item(container, doc_id):
    print('\n1.6 Replace an Item\n')

    read_item = await container.read_item(item=doc_id, partition_key=doc_id)
    read_item['subtotal'] = read_item['subtotal'] + 1
    response = await container.replace_item(item=read_item, body=read_item)

    print('Replaced Item\'s Id is {0}, new subtotal={1}'.format(response['id'], response['subtotal']))


async def replace_item_using_etags(container, doc_id):
    print('\n1.7 Replace an Item using Etags and IfMatch\n')
    # The use of etags and if-match/if-none-match options allows users to run conditional replace operations
    # based on the etag value passed. When using if-match, the request will only succeed if the item's latest etag
    # matches the passed in value. For more on optimistic concurrency control, see the link below:
    # https://learn.microsoft.com/azure/cosmos-db/nosql/database-transactions-optimistic-concurrency

    read_item = await container.read_item(item=doc_id, partition_key=doc_id)
    item_etag = read_item["_etag"]
    read_item['subtotal'] = read_item['subtotal'] + 1
    response = await container.replace_item(
        read_item,
        read_item,
        if_match=item_etag)

    print('Replaced Item\'s Id is {0}, new subtotal={1}'.format(response['id'], response['subtotal']))

    read_item = await container.read_item(item=doc_id, partition_key=doc_id)
    read_item['subtotal'] = read_item['subtotal'] + 1
    response = await container.replace_item(
        read_item,
        read_item,
        if_none_match="some-etag")

    print('Replaced Item\'s Id is {0}, new subtotal={1}'.format(response['id'], response['subtotal']))


async def upsert_item(container, doc_id):
    print('\n1.8 Upserting an item\n')

    read_item = await container.read_item(item=doc_id, partition_key=doc_id)
    read_item['subtotal'] = read_item['subtotal'] + 1
    response = await container.upsert_item(body=read_item)

    print('Upserted Item\'s Id is {0}, new subtotal={1}'.format(response['id'], response['subtotal']))


async def conditional_patch_item(container, doc_id):
    print('\n1.9 Patching Item by Id based on filter\n')
    operations = [
        {"op": "add", "path": "/favorite_color", "value": "red"},
        {"op": "remove", "path": "/ttl"},
        {"op": "replace", "path": "/tax_amount", "value": 14},
        {"op": "set", "path": "/items/0/discount", "value": 20.0512},
        {"op": "incr", "path": "/total_due", "value": 5},
        {"op": "move", "from": "/freight", "path": "/service_addition"}
    ]
    filter_predicate = "from c where c.tax_amount = 10"

    print("Filter predicate match failure will result in BadRequestException.")
    try:
        await container.patch_item(item=doc_id, partition_key=doc_id,
                                   patch_operations=operations, filter_predicate=filter_predicate)
    except exceptions.CosmosHttpResponseError as e:
        assert (e.status_code == StatusCodes.PRECONDITION_FAILED)
        print("Failed as expected.")


async def patch_item(container, doc_id):
    print('\n1.10 Patching Item by Id\n')

    operations = [
        {"op": "add", "path": "/favorite_color", "value": "red"},
        {"op": "remove", "path": "/ttl"},
        {"op": "replace", "path": "/tax_amount", "value": 14},
        {"op": "set", "path": "/items/0/discount", "value": 20.0512},
        {"op": "incr", "path": "/total_due", "value": 5},
        {"op": "move", "from": "/freight", "path": "/service_addition"}
    ]

    response = await container.patch_item(item=doc_id, partition_key=doc_id, patch_operations=operations)
    print('Patched Item\'s Id is {0}, new path favorite color={1}, removed path ttl={2}, replaced path tax_amount={3},'
          ' set path for item at index 0 of discount={4}, increase in path total_due, new total_due={5}, '
          'move from path freight={6} to path service_addition={7}'.format(response["id"], response["favorite_color"],
                                                                           response.get("ttl"), response["tax_amount"],
                                                                           response["items"][0].get("discount"),
                                                                           response["total_due"],
                                                                           response.get("freight"),
                                                                           response["service_addition"]))


async def execute_item_batch(database):
    print('\n1.11 Executing Batch Item operations\n')
    container = await database.create_container_if_not_exists(id="batch_container",
                                                              partition_key=PartitionKey(path='/account_number'))
    # We create three items to use for the sample.
    await container.create_item(get_sales_order("read_item"))
    await container.create_item(get_sales_order("delete_item"))
    create_response = await container.create_item(get_sales_order("replace_item"))

    # We create our batch operations
    create_item_operation = ("create", (get_sales_order("create_item"),))
    upsert_item_operation = ("upsert", (get_sales_order("upsert_item"),))
    read_item_operation = ("read", ("read_item",))
    delete_item_operation = ("delete", ("delete_item",))
    replace_item_operation = ("replace", ("replace_item", {"id": "replace_item", 'account_number': 'Account1',
                                                           "message": "item was replaced"}))
    replace_item_if_match_operation = ("replace",
                                       ("replace_item", {"id": "replace_item", 'account_number': 'Account1',
                                                         "message": "item was replaced"}),
                                       {"if_match_etag": create_response.get_response_headers().get("etag")})
    replace_item_if_none_match_operation = ("replace",
                                            ("replace_item", {"id": "replace_item", 'account_number': 'Account1',
                                                              "message": "item was replaced"}),
                                            {"if_none_match_etag": create_response.get_response_headers().get("etag")})

    # Put our operations into a list
    batch_operations = [
        create_item_operation,
        upsert_item_operation,
        read_item_operation,
        delete_item_operation,
        replace_item_operation,
        replace_item_if_match_operation,
        replace_item_if_none_match_operation]

    # Run that list of operations
    batch_results = await container.execute_item_batch(batch_operations=batch_operations, partition_key="Account1")
    # Batch results are returned as a list of item operation results - or raise a CosmosBatchOperationError if
    # one of the operations failed within your batch request.
    print("\nResults for the batch operations: {}\n".format(batch_results))

    # You can also use this logic to read directly from a file into the batch you'd like to create:
    with open("file_name.txt", "r") as data_file:
        container.execute_item_batch([("upsert", (t,)) for t in data_file.readlines()])

    # For error handling, you should use try/ except with CosmosBatchOperationError and use the information in the
    # error returned for your application debugging, making it easy to pinpoint the failing operation
    # [START handle_batch_error]
    batch_operations = [create_item_operation, create_item_operation]
    try:
        await container.execute_item_batch(batch_operations, partition_key="Account1")
    except exceptions.CosmosBatchOperationError as e:
        error_operation_index = e.error_index
        error_operation_response = e.operation_responses[error_operation_index]
        error_operation = batch_operations[error_operation_index]
        print("\nError operation: {}, error operation response: {}\n".format(error_operation, error_operation_response))
    # [END handle_batch_error]


async def delete_item(container, doc_id):
    print('\n1.12 Deleting Item by Id\n')

    await container.delete_item(item=doc_id, partition_key=doc_id)

    print('Deleted item\'s Id is {0}'.format(doc_id))


async def delete_all_items_by_partition_key(db, partitionkey):
    print('\n1.13 Deleting all Items by Partition Key\n')

    # A container with a partition key that is different from id is needed
    container = await db.create_container_if_not_exists(id="Partition Key Delete Container",
                                                        partition_key=PartitionKey(path='/company'))
    sales_order_company_A1 = get_sales_order("SalesOrderCompanyA1")
    sales_order_company_A1["company"] = partitionkey
    await container.upsert_item(sales_order_company_A1)

    print("\nUpserted Item is {} with Partition Key: {}".format(sales_order_company_A1["id"], partitionkey))

    sales_order_company_A2 = get_sales_order("SalesOrderCompanyA2")
    sales_order_company_A2["company"] = partitionkey
    await container.upsert_item(sales_order_company_A2)

    print("\nUpserted Item is {} with Partition Key: {}".format(sales_order_company_A2["id"], partitionkey))

    sales_order_company_B1 = get_sales_order("SalesOrderCompanyB1")
    sales_order_company_B1["company"] = "companyB"
    await container.upsert_item(sales_order_company_B1)

    print("\nUpserted Item is {} with Partition Key: {}".format(sales_order_company_B1["id"], "companyB"))

    item_list = [item async for item in container.read_all_items()]

    print('Found {0} items'.format(item_list.__len__()))

    for doc in item_list:
        print('Item Id: {0}; Partition Key: {1}'.format(doc.get('id'), doc.get("company")))

    print("\nDelete all items for Partition Key: {}\n".format(partitionkey))

    await container.delete_all_items_by_partition_key(partitionkey)
    item_list = [item async for item in container.read_all_items()]

    print('Found {0} items'.format(item_list.__len__()))

    for doc in item_list:
        print('Item Id: {0}; Partition Key: {1}'.format(doc.get('id'), doc.get("company")))


async def create_mh_items(container):
    print('Creating Items')
    print('\n2.1 Create Item with Multi Hash Partition Key\n')

    # Create a SalesOrder object. This object has nested properties and various types including numbers, DateTimes and strings.
    # This can be saved as JSON as is without converting into rows/columns.
    sales_order = get_sales_order("SalesOrder1")
    await container.create_item(body=sales_order)

    # As your app evolves, let's say your object has a new schema. You can insert SalesOrderV2 objects without any
    # changes to the database tier.
    sales_order2 = get_sales_order_v2("SalesOrder2")
    await container.create_item(body=sales_order2)


async def read_mh_item(container, doc_id, pk):
    print('\n2.2 Reading Item by Multi Hash Partition Key\n')

    # Note that Reads require a partition key to be specified.
    response = await container.read_item(item=doc_id, partition_key=pk)

    print('Item read by Partition Key {0}'.format(pk))
    print('Account Number: {0}'.format(response.get('account_number')))
    print('Purchase Order Number: {0}'.format(response.get('purchase_order_number')))


async def query_mh_items(container, pk):
    print('\n2.3 Querying for an  Item by Multi Hash Partition Key\n')

    query_items_response = container.query_items(
        query="SELECT * FROM r WHERE r.account_number=@account_number and r.purchase_order_number=@purchase_order_number",
        parameters=[
            {"name": "@account_number", "value": pk[0]},
            {"name": "@purchase_order_number", "value": pk[1]}
        ]
    )

    items = [item async for item in query_items_response]

    print('Account Number: {0}'.format(items[0].get('account_number')))
    print('Purchase Order Number: {0}'.format(items[0].get('purchase_order_number')))


async def replace_mh_item(container, doc_id, pk):
    print('\n2.4 Replace an Item with Multi Hash Partition Key\n')

    read_item = await container.read_item(item=doc_id, partition_key=pk)
    read_item['subtotal'] = read_item['subtotal'] + 1
    response = await container.replace_item(item=read_item, body=read_item)

    print('Replaced Item\'s Account Number is {0}, Purchase Order Number is {1}, new subtotal={2}'.format(
        response['account_number'], response['purchase_order_number'], response['subtotal']))


async def upsert_mh_item(container, doc_id, pk):
    print('\n2.5 Upserting an item with Multi Hash Partition Key\n')

    read_item = await container.read_item(item=doc_id, partition_key=pk)
    read_item['subtotal'] = read_item['subtotal'] + 1
    response = await container.upsert_item(body=read_item)

    print('Replaced Item\'s Account Number is {0}, Purchase Order Number is {1}, new subtotal={2}'.format(
        response['account_number'], response['purchase_order_number'], response['subtotal']))


async def patch_mh_item(container, doc_id, pk):
    print('\n2.6 Patching Item by Multi Hash Partition Key\n')

    operations = [
        {"op": "add", "path": "/favorite_color", "value": "red"},
        {"op": "remove", "path": "/ttl"},
        {"op": "replace", "path": "/tax_amount", "value": 14},
        {"op": "set", "path": "/items/0/discount", "value": 20.0512},
        {"op": "incr", "path": "/total_due", "value": 5},
        {"op": "move", "from": "/freight", "path": "/service_addition"}
    ]

    response = await container.patch_item(item=doc_id, partition_key=pk, patch_operations=operations)
    print('Patched Item\'s Id is {0}, new path favorite color={1}, removed path ttl={2}, replaced path tax_amount={3},'
          ' set path for item at index 0 of discount={4}, increase in path total_due, new total_due={5}, move from path freight={6}'
          ' to path service_addition={7}'.format(response["id"], response["favorite_color"], response.get("ttl"),
                                                 response["tax_amount"], response["items"][0].get("discount"),
                                                 response["total_due"], response.get("freight"),
                                                 response["service_addition"]))


async def delete_mh_item(container, doc_id, pk):
    print('\n2.7 Deleting Item by Multi Hash Partition Key\n')

    response = await container.delete_item(item=doc_id, partition_key=pk)
    print('Deleted item\'s Account Number is {0} Purchase Order Number is {1}'.format(pk[0], pk[1]))


async def delete_all_items_by_partition_key_mh(db, partitionkey):
    print('\n2.8 Deleting all Items by Partition Key Multi Hash\n')

    # A container with a partition key that is different from id is needed
    container = await db.create_container_if_not_exists(id="Partition Key Delete Container Multi Hash",
                                                        partition_key=PartitionKey(path=['/id', '/company'],
                                                                                   kind='MultiHash'))
    sales_order_company_A1 = get_sales_order(partitionkey[0])
    sales_order_company_A1["company"] = partitionkey[1]
    await container.upsert_item(sales_order_company_A1)

    print("\nUpserted Item is {} with Partition Key: {}".format(sales_order_company_A1["id"], partitionkey))

    sales_order_company_A2 = get_sales_order(partitionkey[0])
    sales_order_company_A2["company"] = partitionkey[1]
    await container.upsert_item(sales_order_company_A2)

    print("\nUpserted Item is {} with Partition Key: {}".format(sales_order_company_A2["id"], partitionkey))

    sales_order_company_B1 = get_sales_order("SalesOrderCompanyB1")
    sales_order_company_B1["company"] = "companyB"
    await container.upsert_item(sales_order_company_B1)

    print("\nUpserted Item is {} with Partition Key: {}".format(sales_order_company_B1["id"], "companyB"))

    item_list = [item async for item in container.read_all_items()]

    print('Found {0} items'.format(item_list.__len__()))

    for doc in item_list:
        print('Item Id: {0}; Partition Key: {1}'.format(doc.get('id'), doc.get("company")))

    print("\nDelete all items for Partition Key: {}\n".format(partitionkey))

    await container.delete_all_items_by_partition_key(partitionkey)
    item_list = [item async for item in container.read_all_items()]

    print('Found {0} items'.format(item_list.__len__()))

    for doc in item_list:
        print('Item Id: {0}; Partition Key: {1}'.format(doc.get('id'), doc.get("company")))


async def query_items_with_continuation_token_size_limit(container, doc_id):
    print('\n2.9 Query Items With Continuation Token Size Limit.\n')

    size_limit_in_kb = 8
    sales_order = get_sales_order(doc_id)
    await container.create_item(body=sales_order)

    # set continuation_token_limit to 8 to limit size to 8KB
    items = container.query_items(
        query="SELECT * FROM r",
        partition_key=doc_id,
        continuation_token_limit=size_limit_in_kb
    )

    print('Continuation Token size has been limited to {}KB.'.format(size_limit_in_kb))


def get_sales_order(item_id):
    order1 = {'id': item_id,
              'account_number': 'Account1',
              'purchase_order_number': 'PO18009186470',
              'order_date': datetime.date(2005, 1, 10).strftime('%c'),
              'subtotal': 419.4589,
              'tax_amount': 12.5838,
              'freight': 472.3108,
              'total_due': 985.018,
              'items': [
                  {'order_qty': 1,
                   'product_id': 100,
                   'unit_price': 418.4589,
                   'line_price': 418.4589
                   }
              ],
              'ttl': 60 * 60 * 24 * 30
              }

    return order1


def get_sales_order_v2(item_id):
    # notice new fields have been added to the sales order
    order2 = {'id': item_id,
              'account_number': 'Account2',
              'purchase_order_number': 'PO15428132599',
              'order_date': datetime.date(2005, 7, 11).strftime('%c'),
              'due_date': datetime.date(2005, 7, 21).strftime('%c'),
              'shipped_date': datetime.date(2005, 7, 15).strftime('%c'),
              'subtotal': 6107.0820,
              'tax_amount': 586.1203,
              'freight': 183.1626,
              'discount_amt': 1982.872,
              'total_due': 4893.3929,
              'items': [
                  {'order_qty': 3,
                   'product_code': 'A-123',  # notice how in item details we no longer reference a ProductId
                   'product_name': 'Product 1',  # instead we have decided to denormalise our schema and include
                   'currency_symbol': '$',  # the Product details relevant to the Order on to the Order directly
                   'currency_code': 'USD',  # this is a typical refactor that happens in the course of an application
                   'unit_price': 17.1,  # that would have previously required schema changes and data migrations etc.
                   'line_price': 5.7
                   }
              ],
              'ttl': 60 * 60 * 24 * 30
              }

    return order2


async def run_sample():
    async with CosmosClient(HOST, {'masterKey': MASTER_KEY}) as client:
        try:
            # setup database for this sample
            db = await client.create_database_if_not_exists(id=DATABASE_ID)

            # setup container for this sample
            container = await db.create_container_if_not_exists(id=CONTAINER_ID,
                                                                partition_key=PartitionKey(path='/id', kind='Hash'))
            print('Container with id \'{0}\' created'.format(CONTAINER_ID))

            await create_items(container)
            await read_item(container, 'SalesOrder1')
            await read_items(container)
            await query_items(container, 'SalesOrder1')
            await query_items_with_continuation_token(container)
            await query_items_with_feed_ranges_and_pagination(container)
            await query_items_single_partition_with_pagination(container)
            await query_items_cross_partition_with_pagination(container)
            await replace_item(container, 'SalesOrder1')
            await replace_item_using_etags(container, 'SalesOrder1')
            await upsert_item(container, 'SalesOrder1')
            await conditional_patch_item(container, 'SalesOrder1')
            await patch_item(container, 'SalesOrder1')
            await delete_item(container, 'SalesOrder1')
            await delete_all_items_by_partition_key(db, "CompanyA")
            await query_items_with_continuation_token_size_limit(container, 'SalesOrder1')

            # setup MultiHash samples
            container_multi_hash = await db.create_container_if_not_exists(id=CONTAINER_MH_ID,
                                                                           partition_key=PartitionKey(
                                                                               path=['/account_number',
                                                                                     '/purchase_order_number'],
                                                                               kind="MultiHash"))

            await create_mh_items(container_multi_hash)
            await read_mh_item(container_multi_hash, 'SalesOrder1', ['Account1', 'PO18009186470'])
            await query_mh_items(container_multi_hash, ['Account1', 'PO18009186470'])
            await replace_mh_item(container_multi_hash, 'SalesOrder1', ['Account1', 'PO18009186470'])
            await upsert_mh_item(container_multi_hash, 'SalesOrder1', ['Account1', 'PO18009186470'])
            await patch_mh_item(container_multi_hash, 'SalesOrder1', ['Account1', 'PO18009186470'])
            await delete_mh_item(container_multi_hash, 'SalesOrder1', ['Account1', 'PO18009186470'])
            await delete_all_items_by_partition_key_mh(db, ["SalesOrderCompany", "CompanyA"])

            # cleanup database after sample
            try:
                await client.delete_database(db)

            except exceptions.CosmosResourceNotFoundError:
                pass

        except exceptions.CosmosHttpResponseError as e:
            print('\nrun_sample has caught an error. {0}'.format(e.message))

        finally:
            print("\nrun_sample done")


if __name__ == '__main__':
    asyncio.run(run_sample())



from Configurations import Configurations
from ConflictWorker import ConflictWorker
from Worker import Worker
from multiprocessing.pool import ThreadPool
import azure.cosmos.documents as documents
from azure.cosmos import CosmosClient

class MultiMasterScenario(object):
    def __init__(self):
        self.account_endpoint = Configurations.ENDPOINT
        self.account_key = Configurations.ACCOUNT_KEY

        self.regions = Configurations.REGIONS.split(';')

        self.database_name = Configurations.DATABASE_NAME
        self.manual_collection_name = Configurations.MANUAL_COLLECTION_NAME
        self.lww_collection_name = Configurations.LWW_COLLECTION_NAME
        self.udp_collection_name = Configurations.UDP_COLLECTION_NAME
        self.basic_collection_name = Configurations.BASIC_COLLECTION_NAME

        self.workers = []
        self.conflict_worker = ConflictWorker(self.database_name, self.basic_collection_name, self.manual_collection_name, self.lww_collection_name, self.udp_collection_name)
        self.pool = ThreadPool(processes = len(self.regions))

        for region in self.regions:
            connection_policy = documents.ConnectionPolicy()
            connection_policy.UseMultipleWriteLocations = True
            connection_policy.PreferredLocations = [region]

            client = CosmosClient(
                url=self.account_endpoint,
                credential=self.account_key,
                consistency_level=documents.ConsistencyLevel.Session,
                connection_policy=connection_policy)

            self.workers.append(Worker(client, self.database_name, self.basic_collection_name))

            self.conflict_worker.add_client(client)

    def initialize_async(self):
        self.conflict_worker.initialize_async()
        print("Initialized collections.")

    def run_basic_async(self):
        print("\n####################################################")
        print("Basic Active-Active")
        print("####################################################")

        print("1) Starting insert loops across multiple regions ...")

        documents_to_insert_per_worker = 100

        run_loop_futures = []
        for worker in self.workers:
            run_loop_future = self.pool.apply_async(worker.run_loop_async, (documents_to_insert_per_worker,))
            run_loop_futures.append(run_loop_future)

        for run_loop_future in run_loop_futures:
            run_loop_future.get()

        print("2) Reading from every region ...")

        expected_documents = len(self.workers) * documents_to_insert_per_worker

        read_all_futures = []
        for worker in self.workers:
            read_all_future = self.pool.apply_async(worker.read_all_async, (expected_documents,))
            read_all_futures.append(read_all_future)

        for read_all_future in read_all_futures:
            read_all_future.get()

        print("3) Deleting all the documents ...")

        self.workers[0].delete_all_async()

        print("####################################################")

    def run_manual_conflict_async(self):
        print("\n####################################################")
        print("Manual Conflict Resolution")
        print("####################################################")

        self.conflict_worker.run_manual_conflict_async()
        print("####################################################")

    def run_LWW_async(self):
        print("\n####################################################")
        print("LWW Conflict Resolution")
        print("####################################################")

        self.conflict_worker.run_LWW_conflict_async()
        print("####################################################")

    def run_UDP_async(self):
        print("\n####################################################")
        print("UDP Conflict Resolution")
        print("####################################################")

        self.conflict_worker.run_UDP_async()
        print("####################################################")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import unittest
import uuid

import pytest

from azure.cosmos.aio import CosmosClient
import test_config
from azure.cosmos.exceptions import CosmosHttpResponseError


@pytest.mark.cosmosEmulator
class TestNoneOptionsAsync(unittest.IsolatedAsyncioTestCase):
    configs = test_config.TestConfig
    host = configs.host
    masterKey = configs.masterKey
    connectionPolicy = configs.connectionPolicy

    async def asyncSetUp(self):
        self.client = CosmosClient(self.host, self.masterKey)
        await self.client.__aenter__()
        self.database = self.client.get_database_client(self.configs.TEST_DATABASE_ID)
        self.container = self.database.get_container_client(self.configs.TEST_SINGLE_PARTITION_CONTAINER_ID)

    async def asyncTearDown(self):
        await self.client.close()

    async def _create_sample_item(self):
        item = {"id": str(uuid.uuid4()), "pk": "pk-value", "value": 42}
        await self.container.create_item(item, pre_trigger_include=None, post_trigger_include=None, indexing_directive=None,
                                     enable_automatic_id_generation=False, session_token=None, initial_headers=None,
                                     priority=None, no_response=None, retry_write=None, throughput_bucket=None)
        return item

    async def test_container_read_none_options_async(self):
        result = await self.container.read(populate_partition_key_range_statistics=None, populate_quota_info=None,
                                      priority=None, initial_headers=None)
        assert result

    async def test_container_create_item_none_options_async(self):
        item = {"id": str(uuid.uuid4()), "pk": "pk-value", "value": 1}
        created = await self.container.create_item(item, pre_trigger_include=None, post_trigger_include=None,
                                               indexing_directive=None, enable_automatic_id_generation=False,
                                               session_token=None, initial_headers=None, priority=None, no_response=None,
                                               retry_write=None, throughput_bucket=None)
        assert created["id"] == item["id"]

    async def test_container_read_item_none_options_async(self):
        item = await self._create_sample_item()
        read_back = await self.container.read_item(item["id"], partition_key=item["pk"], post_trigger_include=None,
                                              session_token=None, initial_headers=None,
                                              max_integrated_cache_staleness_in_ms=None, priority=None,
                                              throughput_bucket=None)
        assert read_back["id"] == item["id"]

    async def test_container_read_all_items_none_options_async(self):
        await self._create_sample_item()
        pager = self.container.read_all_items(max_item_count=None, session_token=None, initial_headers=None,
                                         max_integrated_cache_staleness_in_ms=None, priority=None,
                                         throughput_bucket=None)
        items = [item async for item in pager]
        assert len(items) >= 1

    async def test_container_read_items_none_options_async(self):
        item = await self._create_sample_item()
        results = await self.container.read_items([(item["id"], item["pk"])], max_concurrency=None, consistency_level=None,
                                              session_token=None, initial_headers=None, excluded_locations=None,
                                              priority=None, throughput_bucket=None)
        assert any(r["id"] == item["id"] for r in results)

    async def test_container_query_items_none_options_partition_async(self):
        await self._create_sample_item()
        pager = self.container.query_items("SELECT * FROM c", continuation_token_limit=None, enable_scan_in_query=None,
                                      initial_headers=None, max_integrated_cache_staleness_in_ms=None, max_item_count=None,
                                      parameters=None, partition_key=None, populate_index_metrics=None,
                                      populate_query_metrics=None, priority=None, response_hook=None, session_token=None,
                                      throughput_bucket=None)
        items = [doc async for doc in pager]
        assert len(items) >= 1

    async def test_upsert_item_none_options_async(self):
        item = {"id": str(uuid.uuid4()), "pk": "pk-value", "value": 5}
        upserted = await self.container.upsert_item(item, pre_trigger_include=None, post_trigger_include=None,
                                                    session_token=None, initial_headers=None, etag=None,
                                                    match_condition=None, priority=None, no_response=None,
                                                    retry_write=None, throughput_bucket=None)
        assert upserted["id"] == item["id"]

    async def test_replace_item_none_options_async(self):
        item = await self._create_sample_item()
        new_body = {"id": item["id"], "pk": item["pk"], "value": 999}
        replaced = await self.container.replace_item(item["id"], new_body, pre_trigger_include=None,
                                                     post_trigger_include=None, session_token=None,
                                                     initial_headers=None, etag=None, match_condition=None,
                                                     priority=None, no_response=None, retry_write=None,
                                                     throughput_bucket=None)
        assert replaced["value"] == 999

    async def test_patch_item_none_options_async(self):
        item = await self._create_sample_item()
        operations = [{"op": "add", "path": "/patched", "value": True}]
        patched = await self.container.patch_item(item["id"], partition_key=item["pk"], patch_operations=operations,
                                                  filter_predicate=None, pre_trigger_include=None, post_trigger_include=None,
                                                  session_token=None, etag=None, match_condition=None, priority=None,
                                                  no_response=None, retry_write=None, throughput_bucket=None)
        assert patched["patched"] is True

    async def test_delete_item_none_options_async(self):
        item = await self._create_sample_item()
        await self.container.delete_item(item["id"], partition_key=item["pk"], pre_trigger_include=None,
                                         post_trigger_include=None, session_token=None, initial_headers=None,
                                         etag=None, match_condition=None, priority=None, retry_write=None,
                                         throughput_bucket=None)
        with self.assertRaises(CosmosHttpResponseError):
            await self.container.read_item(item["id"], partition_key=item["pk"], post_trigger_include=None,
                                           session_token=None, initial_headers=None,
                                           max_integrated_cache_staleness_in_ms=None, priority=None,
                                           throughput_bucket=None)

    async def test_get_throughput_none_options_async(self):
        tp = await self.container.get_throughput(response_hook=None)
        assert tp.offer_throughput > 0

    async def test_list_conflicts_none_options_async(self):
        pager = self.container.list_conflicts(max_item_count=None, response_hook=None)
        conflicts = [c async for c in pager]
        assert conflicts == conflicts  # simple sanity (may be empty)

    async def test_query_conflicts_none_options_async(self):
        pager = self.container.query_conflicts("SELECT * FROM c", parameters=None, partition_key=None,
                                               max_item_count=None, response_hook=None)
        conflicts = [c async for c in pager]
        assert conflicts == conflicts

    async def test_delete_all_items_by_partition_key_none_options_async(self):
        pk_value = "delete-pk"
        for _ in range(2):
            item = {"id": str(uuid.uuid4()), "pk": pk_value, "value": 1}
            await self.container.create_item(item, pre_trigger_include=None, post_trigger_include=None,
                                             indexing_directive=None, enable_automatic_id_generation=False,
                                             session_token=None, initial_headers=None, priority=None, no_response=None,
                                             retry_write=None, throughput_bucket=None)
        await self.container.delete_all_items_by_partition_key(pk_value, pre_trigger_include=None,
                                                               post_trigger_include=None, session_token=None,
                                                               throughput_bucket=None)
        # Just ensure query still works with None options
        pager = self.container.query_items("SELECT * FROM c WHERE c.pk = @pk", parameters=[{"name": "@pk", "value": pk_value}],
                                           partition_key=None, continuation_token_limit=None, enable_scan_in_query=None,
                                           initial_headers=None, max_integrated_cache_staleness_in_ms=None, max_item_count=None,
                                           populate_index_metrics=None, populate_query_metrics=None, priority=None,
                                           response_hook=None, session_token=None, throughput_bucket=None)
        _items = [doc async for doc in pager]
        assert _items == _items

    async def test_execute_item_batch_none_options_async(self):
        pk_value = "batch-pk"
        id1 = str(uuid.uuid4())
        id2 = str(uuid.uuid4())
        ops = [
            ("create", ({"id": id1, "pk": pk_value},)),
            ("create", ({"id": id2, "pk": pk_value},)),
        ]
        batch_result = await self.container.execute_item_batch(ops, partition_key=pk_value,
                                                                pre_trigger_include=None, post_trigger_include=None,
                                                                session_token=None, priority=None,
                                                                throughput_bucket=None)
        assert any(r.get("resourceBody").get("id") == id1 for r in batch_result) or any(r.get("resourceBody").get("id") == id2 for r in batch_result)


    async def test_query_items_change_feed_none_options_async(self):
        for _ in range(15):
            await self.container.create_item({"id": str(uuid.uuid4()), "pk": "cf-pk", "value": 100},
                                             pre_trigger_include=None, post_trigger_include=None, indexing_directive=None,
                                             enable_automatic_id_generation=False, session_token=None, initial_headers=None,
                                             priority=None, no_response=None, retry_write=None, throughput_bucket=None)
        # Obtain the change feed pager with all optional parameters set to None (including partition_key)
        pager = self.container.query_items_change_feed(max_item_count=None, start_time="Beginning", partition_key=None,
                                                       priority=None, mode=None, response_hook=None)

        changes = [doc async for doc in pager]
        assert len(changes) >= 1



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

"""Internal class for connection reset retry policy implementation in the Azure
Cosmos database service.
"""
from . import http_constants
from .documents import _OperationType

# pylint: disable=protected-access


class DefaultRetryPolicy(object):

    error_codes = http_constants._ErrorCodes
    CONNECTION_ERROR_CODES = [
        error_codes.WindowsInterruptedFunctionCall,
        error_codes.WindowsFileHandleNotValid,
        error_codes.WindowsPermissionDenied,
        error_codes.WindowsBadAddress,
        error_codes.WindowsInvalidArgumnet,
        error_codes.WindowsResourceTemporarilyUnavailable,
        error_codes.WindowsOperationNowInProgress,
        error_codes.WindowsAddressAlreadyInUse,
        error_codes.WindowsConnectionResetByPeer,
        error_codes.WindowsCannotSendAfterSocketShutdown,
        error_codes.WindowsConnectionTimedOut,
        error_codes.WindowsConnectionRefused,
        error_codes.WindowsNameTooLong,
        error_codes.WindowsHostIsDown,
        error_codes.WindowsNoRouteTohost,
        error_codes.LinuxConnectionReset,
    ]

    def __init__(self, *args):
        self._max_retry_attempt_count = 10
        self.current_retry_attempt_count = 0
        self.retry_after_in_milliseconds = 1000
        self.args = args
        self.request = args[0] if args else None

    def needsRetry(self, error_code):
        if error_code in DefaultRetryPolicy.CONNECTION_ERROR_CODES:
            if self.args:
                if _OperationType.IsReadOnlyOperation(self.request.operation_type):
                    return True
                return False
            return True
        return False

    def ShouldRetry(self, exception):
        """Returns true if the request should retry based on the passed-in exception.

        :param exceptions.CosmosHttpResponseError exception:
        :returns: a boolean stating whether the request should be retried
        :rtype: bool
        """
        if (self.current_retry_attempt_count < self._max_retry_attempt_count) and self.needsRetry(
            exception.status_code
        ):
            self.current_retry_attempt_count += 1
            return True
        return False



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import os
import unittest
import uuid

import pytest

import azure.cosmos.aio._retry_utility_async as retry_utility
import azure.cosmos.exceptions as exceptions
import test_config
from azure.cosmos._execution_context.query_execution_info import _PartitionedQueryExecutionInfo
from azure.cosmos.aio import CosmosClient, DatabaseProxy, ContainerProxy
from azure.cosmos.documents import _DistinctType
from azure.cosmos.exceptions import CosmosHttpResponseError
from azure.cosmos.partition_key import PartitionKey

@pytest.mark.cosmosCircuitBreaker
@pytest.mark.cosmosQuery
class TestQueryCrossPartitionAsync(unittest.IsolatedAsyncioTestCase):
    """Test to ensure escaping of non-ascii characters from partition key"""

    created_db: DatabaseProxy = None
    created_container: ContainerProxy = None
    client: CosmosClient = None
    config = test_config.TestConfig
    host = config.host
    masterKey = config.masterKey
    connectionPolicy = config.connectionPolicy
    TEST_CONTAINER_ID = str(uuid.uuid4())
    TEST_DATABASE_ID = config.TEST_DATABASE_ID

    @classmethod
    def setUpClass(cls):
        if (cls.masterKey == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")

    async def asyncSetUp(self):
        use_multiple_write_locations = False
        if os.environ.get("AZURE_COSMOS_ENABLE_CIRCUIT_BREAKER", "False") == "True":
            use_multiple_write_locations = True
        self.client = CosmosClient(self.host, self.masterKey, multiple_write_locations=use_multiple_write_locations)
        self.created_db = self.client.get_database_client(self.TEST_DATABASE_ID)
        self.created_container = await self.created_db.create_container(
            self.TEST_CONTAINER_ID,
            PartitionKey(path="/pk"),
            offer_throughput=test_config.TestConfig.THROUGHPUT_FOR_5_PARTITIONS)
        if self.host == "https://localhost:8081/":
            os.environ["AZURE_COSMOS_DISABLE_NON_STREAMING_ORDER_BY"] = "True"

    async def asyncTearDown(self):
        try:
            await self.created_db.delete_container(self.TEST_CONTAINER_ID)
        except CosmosHttpResponseError:
            pass
        finally:
            await self.client.close()

    async def test_first_and_last_slashes_trimmed_for_query_string_async(self):
        doc_id = 'myId' + str(uuid.uuid4())
        document_definition = {'pk': 'pk', 'id': doc_id}
        await self.created_container.create_item(body=document_definition)

        query = 'SELECT * from c'
        query_iterable = self.created_container.query_items(
            query=query,
            partition_key='pk'
        )
        iter_list = [item async for item in query_iterable]
        assert iter_list[0]['id'] == doc_id

    async def test_query_change_feed_with_pk_async(self):
        # The test targets partition #3
        partition_key = "pk"

        # Read change feed without passing any options
        query_iterable = self.created_container.query_items_change_feed()
        iter_list = [item async for item in query_iterable]
        assert len(iter_list) == 0

        # Read change feed from current should return an empty list
        query_iterable = self.created_container.query_items_change_feed(partition_key=partition_key)
        iter_list = [item async for item in query_iterable]
        assert len(iter_list) == 0
        if 'Etag' in self.created_container.client_connection.last_response_headers:
            assert self.created_container.client_connection.last_response_headers['Etag'] != ''
        elif 'etag' in self.created_container.client_connection.last_response_headers:
            assert self.created_container.client_connection.last_response_headers['etag'] != ''
        else:
            self.fail("No Etag or etag found in last response headers")

        # Read change feed from beginning should return an empty list
        query_iterable = self.created_container.query_items_change_feed(
            is_start_from_beginning=True,
            partition_key=partition_key
        )
        iter_list = [item async for item in query_iterable]
        assert len(iter_list) == 0
        if 'Etag' in self.created_container.client_connection.last_response_headers:
            continuation1 = self.created_container.client_connection.last_response_headers['Etag']
        elif 'etag' in self.created_container.client_connection.last_response_headers:
            continuation1 = self.created_container.client_connection.last_response_headers['etag']
        else:
            self.fail("No Etag or etag found in last response headers")
        assert continuation1 != ''

        # Create a document. Read change feed should return be able to read that document
        document_definition = {'pk': 'pk', 'id': 'doc1'}
        await self.created_container.create_item(body=document_definition)
        query_iterable = self.created_container.query_items_change_feed(
            is_start_from_beginning=True,
            partition_key=partition_key
        )
        iter_list = [item async for item in query_iterable]
        assert len(iter_list) == 1
        assert iter_list[0]['id'] == 'doc1'
        if 'Etag' in self.created_container.client_connection.last_response_headers:
            continuation2 = self.created_container.client_connection.last_response_headers['Etag']
        elif 'etag' in self.created_container.client_connection.last_response_headers:
            continuation2 = self.created_container.client_connection.last_response_headers['etag']
        else:
            self.fail("No Etag or etag found in last response headers")
        assert continuation2 != ''
        assert continuation2 != continuation1

        # Create two new documents. Verify that change feed contains the 2 new documents
        # with page size 1 and page size 100
        document_definition = {'pk': 'pk', 'id': 'doc2'}
        await self.created_container.create_item(body=document_definition)
        document_definition = {'pk': 'pk', 'id': 'doc3'}
        await self.created_container.create_item(body=document_definition)

        for pageSize in [2, 100]:
            # verify iterator
            query_iterable = self.created_container.query_items_change_feed(
                continuation=continuation2,
                max_item_count=pageSize,
                partition_key=partition_key)
            it = query_iterable.__aiter__()
            expected_ids = 'doc2.doc3.'
            actual_ids = ''
            async for item in it:
                actual_ids += item['id'] + '.'
            assert actual_ids == expected_ids

            # verify by_page
            # the options is not copied, therefore it need to be restored
            query_iterable = self.created_container.query_items_change_feed(
                continuation=continuation2,
                max_item_count=pageSize,
                partition_key=partition_key
            )
            count = 0
            expected_count = 2
            all_fetched_res = []
            pages = query_iterable.by_page()
            async for items in await pages.__anext__():
                count += 1
                all_fetched_res.append(items)
            assert count == expected_count

            actual_ids = ''
            for item in all_fetched_res:
                actual_ids += item['id'] + '.'
            assert actual_ids == expected_ids

        # verify reading change feed from the beginning
        query_iterable = self.created_container.query_items_change_feed(
            is_start_from_beginning=True,
            partition_key=partition_key
        )
        expected_ids = ['doc1', 'doc2', 'doc3']
        it = query_iterable.__aiter__()
        for i in range(0, len(expected_ids)):
            doc = await it.__anext__()
            assert doc['id'] == expected_ids[i]
        if 'Etag' in self.created_container.client_connection.last_response_headers:
            continuation3 = self.created_container.client_connection.last_response_headers['Etag']
        elif 'etag' in self.created_container.client_connection.last_response_headers:
            continuation3 = self.created_container.client_connection.last_response_headers['etag']
        else:
            self.fail("No Etag or etag found in last response headers")

        # verify reading empty change feed
        query_iterable = self.created_container.query_items_change_feed(
            continuation=continuation3,
            is_start_from_beginning=True,
            partition_key=partition_key
        )
        iter_list = [item async for item in query_iterable]
        assert len(iter_list) == 0

    async def test_populate_query_metrics_async(self):
        doc_id = 'MyId' + str(uuid.uuid4())
        document_definition = {'pk': 'pk', 'id': doc_id}
        await self.created_container.create_item(body=document_definition)

        query = 'SELECT * from c'
        query_iterable = self.created_container.query_items(
            query=query,
            partition_key='pk',
            populate_query_metrics=True
        )

        iter_list = [item async for item in query_iterable]
        assert iter_list[0]['id'] == doc_id

        metrics_header_name = 'x-ms-documentdb-query-metrics'
        assert metrics_header_name in self.created_container.client_connection.last_response_headers
        metrics_header = self.created_container.client_connection.last_response_headers[metrics_header_name]
        # Validate header is well-formed: "key1=value1;key2=value2;etc"
        metrics = metrics_header.split(';')
        assert len(metrics) > 1
        assert all(['=' in x for x in metrics])

    async def validate_query_requests_count(self, query_iterable, expected_count):
        self.count = 0
        self.OriginalExecuteFunction = retry_utility.ExecuteFunctionAsync
        retry_utility.ExecuteFunctionAsync = self._mock_execute_function
        item_pages = query_iterable.by_page()
        while True:
            try:
                page = await item_pages.__anext__()
                assert len([item async for item in page]) > 0
            except StopAsyncIteration:
                break
        retry_utility.ExecuteFunctionAsync = self.OriginalExecuteFunction
        assert self.count == expected_count
        self.count = 0

    async def _mock_execute_function(self, function, *args, **kwargs):
        self.count += 1
        return await self.OriginalExecuteFunction(function, *args, **kwargs)

    async def test_get_query_plan_through_gateway_async(self):
        await self._validate_query_plan(query="Select top 10 value count(c.id) from c",
                                        container_link=self.created_container.container_link,
                                        top=10,
                                        order_by=[],
                                        aggregate=['Count'],
                                        select_value=True,
                                        offset=None,
                                        limit=None,
                                        distinct=_DistinctType.NoneType)

        await self._validate_query_plan(query="Select * from c order by c._ts offset 5 limit 10",
                                        container_link=self.created_container.container_link,
                                        top=None,
                                        order_by=['Ascending'],
                                        aggregate=[],
                                        select_value=False,
                                        offset=5,
                                        limit=10,
                                        distinct=_DistinctType.NoneType)

        await self._validate_query_plan(query="Select distinct value c.id from c order by c.id",
                                        container_link=self.created_container.container_link,
                                        top=None,
                                        order_by=['Ascending'],
                                        aggregate=[],
                                        select_value=True,
                                        offset=None,
                                        limit=None,
                                        distinct=_DistinctType.Ordered)

    async def _validate_query_plan(self, query, container_link, top, order_by, aggregate, select_value, offset, limit,
                                   distinct):
        query_plan_dict = await self.client.client_connection._GetQueryPlanThroughGateway(query, container_link)
        query_execution_info = _PartitionedQueryExecutionInfo(query_plan_dict)
        assert query_execution_info.has_rewritten_query()
        assert query_execution_info.has_distinct_type() == (distinct != "None")
        assert query_execution_info.get_distinct_type() == distinct
        assert query_execution_info.has_top() == (top is not None)
        assert query_execution_info.get_top() == top
        assert query_execution_info.has_order_by() == (len(order_by) > 0)
        assert query_execution_info.get_order_by() == order_by
        assert query_execution_info.has_aggregates() == (len(aggregate) > 0)
        assert query_execution_info.get_aggregates() == aggregate
        assert query_execution_info.has_select_value() == select_value
        assert query_execution_info.has_offset() == (offset is not None)
        assert query_execution_info.get_offset() == offset
        assert query_execution_info.has_limit() == (limit is not None)
        assert query_execution_info.get_limit() == limit

    async def test_unsupported_queries_async(self):
        queries = ['SELECT COUNT(1) FROM c', 'SELECT COUNT(1) + 5 FROM c', 'SELECT COUNT(1) + SUM(c) FROM c']
        for query in queries:
            query_iterable = self.created_container.query_items(query=query)
            try:
                results = [item async for item in query_iterable]
                self.fail("query '{}' should have failed".format(query))
            except exceptions.CosmosHttpResponseError as e:
                assert e.status_code == 400

    async def test_query_with_non_overlapping_pk_ranges_async(self):
        query_iterable = self.created_container.query_items("select * from c where c.pk='1' or c.pk='2'")
        assert [item async for item in query_iterable] == []

    async def test_offset_limit_async(self):
        values = []
        for i in range(10):
            document_definition = {'pk': i, 'id': 'myId' + str(uuid.uuid4()), 'value': i // 3}
            current_document = await self.created_container.create_item(body=document_definition)
            values.append(current_document['pk'])

        await self.config._validate_distinct_offset_limit(
            created_collection=self.created_container,
            query='SELECT DISTINCT c["value"] from c ORDER BY c.pk OFFSET 0 LIMIT 2',
            results=[0, 1])

        await self.config._validate_distinct_offset_limit(
            created_collection=self.created_container,
            query='SELECT DISTINCT c["value"] from c ORDER BY c.pk OFFSET 2 LIMIT 2',
            results=[2, 3])

        await self.config._validate_distinct_offset_limit(
            created_collection=self.created_container,
            query='SELECT DISTINCT c["value"] from c ORDER BY c.pk OFFSET 4 LIMIT 3',
            results=[])

        await self.config._validate_offset_limit(created_collection=self.created_container,
                                                 query='SELECT * from c ORDER BY c.pk OFFSET 0 LIMIT 5',
                                                 results=values[:5])

        await self.config._validate_offset_limit(created_collection=self.created_container,
                                                 query='SELECT * from c ORDER BY c.pk OFFSET 5 LIMIT 10',
                                                 results=values[5:])

        await self.config._validate_offset_limit(created_collection=self.created_container,
                                                 query='SELECT * from c ORDER BY c.pk OFFSET 10 LIMIT 5',
                                                 results=[])

        await self.config._validate_offset_limit(created_collection=self.created_container,
                                                 query='SELECT * from c ORDER BY c.pk OFFSET 100 LIMIT 1',
                                                 results=[])

    async def test_distinct_async(self):
        created_database = self.created_db
        distinct_field = 'distinct_field'
        pk_field = "pk"
        different_field = "different_field"

        created_collection = await created_database.create_container(
            id='collection with composite index ' + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk", kind="Hash"),
            indexing_policy={
                "compositeIndexes": [
                    [{"path": "/" + pk_field, "order": "ascending"},
                     {"path": "/" + distinct_field, "order": "ascending"}],
                    [{"path": "/" + distinct_field, "order": "ascending"},
                     {"path": "/" + pk_field, "order": "ascending"}]
                ]
            }
        )
        documents = []
        for i in range(5):
            j = i
            while j > i - 5:
                document_definition = {pk_field: i, 'id': str(uuid.uuid4()), distinct_field: j}
                documents.append(await created_collection.create_item(body=document_definition))
                document_definition = {pk_field: i, 'id': str(uuid.uuid4()), distinct_field: j}
                documents.append(await created_collection.create_item(body=document_definition))
                document_definition = {pk_field: i, 'id': str(uuid.uuid4())}
                documents.append(await created_collection.create_item(body=document_definition))
                j -= 1

        padded_docs = self.config._pad_with_none(documents, distinct_field)

        await self.config._validate_distinct(created_collection=created_collection,  # returns {} and is right number
                                             query='SELECT distinct c.%s from c' % distinct_field,  # nosec
                                             results=self.config._get_distinct_docs(padded_docs, distinct_field, None,
                                                                                    False),
                                             is_select=True,
                                             fields=[distinct_field])

        await self.config._validate_distinct(created_collection=created_collection,
                                             query='SELECT distinct c.%s, c.%s from c' % (distinct_field, pk_field),
                                             # nosec
                                             results=self.config._get_distinct_docs(padded_docs, distinct_field,
                                                                                    pk_field, False),
                                             is_select=True,
                                             fields=[distinct_field, pk_field])

        await self.config._validate_distinct(created_collection=created_collection,
                                             query='SELECT distinct value c.%s from c' % distinct_field,  # nosec
                                             results=self.config._get_distinct_docs(padded_docs, distinct_field, None,
                                                                                    True),
                                             is_select=True,
                                             fields=[distinct_field])

        await self.config._validate_distinct(created_collection=created_collection,
                                             query='SELECT distinct c.%s from c' % different_field,  # nosec
                                             results=['None'],
                                             is_select=True,
                                             fields=[different_field])

        await created_database.delete_container(created_collection.id)

    async def test_distinct_on_different_types_and_field_orders_async(self):
        payloads = [
            {'id': str(uuid.uuid4()), 'f1': 1, 'f2': 'value', 'f3': 100000000000000000, 'f4': [1, 2, '3'],
             'f5': {'f6': {'f7': 2}}},
            {'id': str(uuid.uuid4()), 'f2': '\'value', 'f4': [1.0, 2, '3'], 'f5': {'f6': {'f7': 2.0}}, 'f1': 1.0,
             'f3': 100000000000000000.00},
            {'id': str(uuid.uuid4()), 'f3': 100000000000000000.0, 'f5': {'f6': {'f7': 2}}, 'f2': '\'value', 'f1': 1,
             'f4': [1, 2.0, '3']}
        ]
        for pay in payloads:
            await self.created_container.create_item(pay)

        await self.config._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct value c.f1 from c",
            expected_results=[1]
        )

        await self.config._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct value c.f2 from c",
            expected_results=['value', '\'value']
        )

        await self.config._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct value c.f2 from c order by c.f2",
            expected_results=['value', '\'value']
        )

        await self.config._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct value c.f3 from c",
            expected_results=[100000000000000000]
        )

        await self.config._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct value c.f4 from c",
            expected_results=[[1, 2, '3']]
        )

        await self.config._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct value c.f5.f6 from c",
            expected_results=[{'f7': 2}]
        )

        await self.config._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct c.f1, c.f2, c.f3 from c",
            expected_results=[{'f1': 1, 'f2': 'value', 'f3': 100000000000000000},
                              {'f1': 1.0, 'f2': '\'value', 'f3': 100000000000000000.00}]
        )

        await self.config._validate_distinct_on_different_types_and_field_orders(
            collection=self.created_container,
            query="Select distinct c.f1, c.f2, c.f3 from c order by c.f1",
            expected_results=[{'f1': 1, 'f2': 'value', 'f3': 100000000000000000},
                              {'f1': 1.0, 'f2': '\'value', 'f3': 100000000000000000.00}]
        )

    async def test_paging_with_continuation_token_async(self):
        document_definition = {'pk': 'pk', 'id': '1'}
        await self.created_container.upsert_item(body=document_definition)
        document_definition = {'pk': 'pk', 'id': '2'}
        await self.created_container.upsert_item(body=document_definition)

        query = 'SELECT * from c'
        query_iterable = self.created_container.query_items(
            query=query,
            partition_key='pk',
            max_item_count=1
        )
        pager = query_iterable.by_page()
        await pager.__anext__()
        token = pager.continuation_token

        second_page = [item async for item in await pager.__anext__()][0]

        pager = query_iterable.by_page(token)
        second_page_fetched_with_continuation_token = [item async for item in await pager.__anext__()][0]

        assert second_page['id'] == second_page_fetched_with_continuation_token['id']

    async def test_cross_partition_query_with_continuation_token_async(self):
        document_definition = {'pk': 'pk1', 'id': str(uuid.uuid4())}
        await self.created_container.create_item(body=document_definition)
        document_definition = {'pk': 'pk2', 'id': str(uuid.uuid4())}
        await self.created_container.create_item(body=document_definition)

        query = 'SELECT * from c'
        query_iterable = self.created_container.query_items(
            query=query,
            max_item_count=1)
        pager = query_iterable.by_page()
        await pager.__anext__()
        token = pager.continuation_token
        second_page = [item async for item in await pager.__anext__()][0]

        pager = query_iterable.by_page(token)
        second_page_fetched_with_continuation_token = [item async for item in await pager.__anext__()][0]

        assert second_page['id'] == second_page_fetched_with_continuation_token['id']

    async def test_value_max_query_async(self):
        await self.created_container.create_item(
            {"id": str(uuid.uuid4()), "isComplete": True, "version": 3, "lookupVersion": "console_version"})
        await self.created_container.create_item(
            {"id": str(uuid.uuid4()), "isComplete": True, "version": 2, "lookupVersion": "console_version"})
        query = "Select value max(c.version) FROM c where c.isComplete = true and c.lookupVersion = @lookupVersion"
        query_results = self.created_container.query_items(query, parameters=[
            {"name": "@lookupVersion", "value": "console_version"}
        ])
        item_list = [item async for item in query_results]
        assert len(item_list) == 1
        assert item_list[0] == 3

    async def test_continuation_token_size_limit_query_async(self):
        for i in range(1, 1000):
            await self.created_container.create_item(body=dict(pk='123', id=str(i), some_value=str(i % 3)))
        query = "Select * from c where c.some_value='2'"
        print("Created 1000 items")
        response_query = self.created_container.query_items(query, partition_key='123', max_item_count=100,
                                                            continuation_token_limit=1)
        pager = response_query.by_page()
        await pager.__anext__()
        token = pager.continuation_token
        # Continuation token size should be below 1kb
        assert len(token.encode('utf-8')) <= 1024
        await pager.__anext__()
        token = pager.continuation_token

        # verify a second time
        assert len(token.encode('utf-8')) <= 1024
        print("Test done")

    async def test_cross_partition_query_response_hook_async(self):
        created_collection = await self.created_db.create_container_if_not_exists(
            id="query_response_hook_test" + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk"),
            offer_throughput=12000
        )
        items = [
            {'id': str(uuid.uuid4()), 'pk': '0', 'val': 5},
            {'id': str(uuid.uuid4()), 'pk': '1', 'val': 10},
            {'id': str(uuid.uuid4()), 'pk': '0', 'val': 5},
            {'id': str(uuid.uuid4()), 'pk': '1', 'val': 10},
            {'id': str(uuid.uuid4()), 'pk': '0', 'val': 5},
            {'id': str(uuid.uuid4()), 'pk': '1', 'val': 10}
        ]

        for item in items:
            await created_collection.create_item(body=item)

        response_hook = test_config.ResponseHookCaller()
        item_list = [item async for item in created_collection.query_items("select * from c", response_hook=response_hook)]
        assert len(item_list) == 6
        assert response_hook.count == 2
        await self.created_db.delete_container(created_collection.id)

    async def test_cross_partition_query_pagination_with_max_item_count_async(self):
        """Test cross-partition pagination showing per-page limits and total results."""
        created_collection = await self.created_db.create_container(
            "cross_partition_pagination_test_" + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk"),
            offer_throughput=test_config.TestConfig.THROUGHPUT_FOR_5_PARTITIONS)
        
        # Create 30 items across 3 different partitions
        total_items = 30
        items_per_partition = 10
        partitions = ["pk_1", "pk_2", "pk_3"]
        
        for pk in partitions:
            for i in range(items_per_partition):
                document_definition = {
                    'pk': pk,
                    'id': f'{pk}_item_{i}',
                    'value': i
                }
                await created_collection.create_item(body=document_definition)
        
        # Test cross-partition query with max_item_count
        max_items_per_page = 8
        query = "SELECT * FROM c ORDER BY c['value']"
        query_iterable = created_collection.query_items(
            query=query,
            max_item_count=max_items_per_page
        )
        
        # Iterate through pages and verify per-page counts
        all_fetched_results = []
        page_count = 0
        item_pages = query_iterable.by_page()
        
        async for page in item_pages:
            page_count += 1
            items_in_page = [item async for item in page]
            all_fetched_results.extend(items_in_page)
            
            # Each page should have at most max_item_count items
            # In cross-partition queries, pages may have fewer items
            assert len(items_in_page) <= max_items_per_page
            assert len(items_in_page) > 0  # Pages should not be empty
        
        # Verify total results match expected count
        assert len(all_fetched_results) == total_items
        
        # Verify we got multiple pages
        assert page_count > 1
        
        await self.created_db.delete_container(created_collection.id)
    
    async def test_cross_partition_query_pagination_counting_results_async(self):
        """Test counting total results while paginating across partitions."""
        created_collection = await self.created_db.create_container(
            "cross_partition_count_test_" + str(uuid.uuid4()),
            partition_key=PartitionKey(path="/pk"),
            offer_throughput=test_config.TestConfig.THROUGHPUT_FOR_5_PARTITIONS)
        
        # Create items across multiple partitions with different counts
        partitions_config = [
            ("partition_a", 5),
            ("partition_b", 8),
            ("partition_c", 3),
            ("partition_d", 12)
        ]
        
        total_expected = 0
        for pk, count in partitions_config:
            for i in range(count):
                document_definition = {
                    'pk': pk,
                    'id': f'{pk}_item_{i}',
                    'name': f'Item {i} in {pk}'
                }
                await created_collection.create_item(body=document_definition)
                total_expected += 1
        
        # Query across partitions with pagination
        max_items_per_page = 5
        query = "SELECT * FROM c"
        query_iterable = created_collection.query_items(
            query=query,
            max_item_count=max_items_per_page
        )
        
        # Count items across all pages
        total_count = 0
        page_count = 0
        page_sizes = []
        
        item_pages = query_iterable.by_page()
        async for page in item_pages:
            page_count += 1
            items = [item async for item in page]
            page_size = len(items)
            page_sizes.append(page_size)
            total_count += page_size
            
            # Verify page size constraints
            assert page_size <= max_items_per_page
            assert page_size > 0
        
        # Verify total count matches what we inserted
        assert total_count == total_expected
        
        # Verify we processed multiple pages
        assert page_count > 1
        
        await self.created_db.delete_container(created_collection.id)

if __name__ == '__main__':
    unittest.main()



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Internal class for change feed continuation token by feed range in the Azure Cosmos
database service.
"""
from collections import deque
from typing import Any, Deque, Optional

from azure.cosmos._change_feed.composite_continuation_token import CompositeContinuationToken
from azure.cosmos._change_feed.feed_range_internal import (FeedRangeInternal, FeedRangeInternalEpk,
                                                           FeedRangeInternalPartitionKey)
from azure.cosmos._routing.routing_map_provider import SmartRoutingMapProvider
from azure.cosmos._routing.aio.routing_map_provider import SmartRoutingMapProvider as AsyncSmartRoutingMapProvider
from azure.cosmos._routing.routing_range import Range

class FeedRangeCompositeContinuation:
    _version_property_name = "v"
    _container_rid_property_name = "rid"
    _continuation_property_name = "continuation"

    def __init__(
            self,
            container_rid: str,
            feed_range: FeedRangeInternal,
            continuation: Deque[CompositeContinuationToken]) -> None:
        if container_rid is None:
            raise ValueError("container_rid is missing")

        self._container_rid = container_rid
        self._feed_range = feed_range
        self._continuation = continuation
        self._current_token = self._continuation[0]
        self._initial_no_result_range: Optional[Range] = None

    @property
    def current_token(self) -> CompositeContinuationToken:
        return self._current_token

    def to_dict(self) -> dict[str, Any]:
        json_data = {
            self._version_property_name: "v2",
            self._container_rid_property_name: self._container_rid,
            self._continuation_property_name: [childToken.to_dict() for childToken in self._continuation],
        }
        json_data.update(self._feed_range.to_dict())
        return json_data

    @classmethod
    def from_json(cls, data) -> 'FeedRangeCompositeContinuation':
        version = data.get(cls._version_property_name)
        if version is None:
            raise ValueError(f"Invalid feed range composite continuation token [Missing {cls._version_property_name}]")
        if version != "v2":
            raise ValueError("Invalid feed range composite continuation token [Invalid version]")

        container_rid = data.get(cls._container_rid_property_name)
        if container_rid is None:
            raise ValueError(f"Invalid feed range composite continuation token "
                             f"[Missing {cls._container_rid_property_name}]")

        continuation_data = data.get(cls._continuation_property_name)
        if continuation_data is None:
            raise ValueError(f"Invalid feed range composite continuation token "
                             f"[Missing {cls._continuation_property_name}]")
        if not isinstance(continuation_data, list) or len(continuation_data) == 0:
            raise ValueError(f"Invalid feed range composite continuation token "
                             f"[The {cls._continuation_property_name} must be non-empty array]")
        continuation = [CompositeContinuationToken.from_json(child_range_continuation_token)
                        for child_range_continuation_token in continuation_data]

        # parsing feed range
        feed_range: Optional[FeedRangeInternal] = None
        if data.get(FeedRangeInternalEpk.type_property_name):
            feed_range = FeedRangeInternalEpk.from_json(data)
        elif data.get(FeedRangeInternalPartitionKey.type_property_name):
            feed_range = FeedRangeInternalPartitionKey.from_json(data, continuation[0].feed_range)
        else:
            raise ValueError("Invalid feed range composite continuation token [Missing feed range scope]")

        return cls(container_rid=container_rid, feed_range=feed_range, continuation=deque(continuation))

    def handle_feed_range_gone(
            self,
            routing_provider: SmartRoutingMapProvider,
            collection_link: str,
            feed_options: Optional[dict[str, Any]] = None) -> None:
        overlapping_ranges = routing_provider.get_overlapping_ranges(collection_link,
                                                                     [self._current_token.feed_range], feed_options)

        if len(overlapping_ranges) == 1:
            # merge,reusing the existing the feedRange and continuationToken
            pass
        else:
            # split, remove the parent range and then add new child ranges.
            # For each new child range, using the continuation token from the parent
            self._continuation.popleft()
            for child_range in overlapping_ranges:
                self._continuation.append(
                    CompositeContinuationToken(
                        Range.PartitionKeyRangeToRange(child_range),
                        self._current_token.token))

            self._current_token = self._continuation[0]

    async def handle_feed_range_gone_async(
            self,
            routing_provider: AsyncSmartRoutingMapProvider,
            collection_link: str,
            feed_options: Optional[dict[str, Any]] = None) -> None:
        overlapping_ranges = \
            await routing_provider.get_overlapping_ranges(
                collection_link,
                [self._current_token.feed_range],
                feed_options)

        if len(overlapping_ranges) == 1:
            # merge,reusing the existing the feedRange and continuationToken
            pass
        else:
            # split, remove the parent range and then add new child ranges.
            # For each new child range, using the continuation token from the parent
            self._continuation.popleft()
            for child_range in overlapping_ranges:
                self._continuation.append(
                    CompositeContinuationToken(
                        Range.PartitionKeyRangeToRange(child_range),
                        self._current_token.token))

            self._current_token = self._continuation[0]

    def should_retry_on_not_modified_response(self) -> bool:
        # when getting 304(Not Modified) response from one sub feed range,
        # we will try to fetch for the next sub feed range
        # we will repeat the above logic until we have looped through all sub feed ranges

        # TODO: validate the response headers, can we get the status code
        if len(self._continuation) > 1:
            return self._current_token.feed_range != self._initial_no_result_range

        return False

    def _move_to_next_token(self) -> None:
        first_composition_token = self._continuation.popleft()
        # add the composition token to the end of the list
        self._continuation.append(first_composition_token)
        self._current_token = self._continuation[0]

    def apply_server_response_continuation(self, etag, has_modified_response: bool) -> None:
        self._current_token.update_token(etag)
        if has_modified_response:
            self._initial_no_result_range = None
        else:
            self.apply_not_modified_response()

    def apply_not_modified_response(self) -> None:
        if self._initial_no_result_range is None:
            self._initial_no_result_range = self._current_token.feed_range

    @property
    def feed_range(self) -> FeedRangeInternal:
        return self._feed_range



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Internal class for change feed start from implementation in the Azure Cosmos database service.
"""

from abc import ABC, abstractmethod
from datetime import datetime, timezone
from enum import Enum
from typing import Optional, Union, Literal, Any

from azure.cosmos import http_constants
from azure.cosmos._routing.routing_range import Range

class ChangeFeedStartFromType(Enum):
    BEGINNING = "Beginning"
    NOW = "Now"
    LEASE = "Lease"
    POINT_IN_TIME = "PointInTime"

class ChangeFeedStartFromInternal(ABC):
    """Abstract class for change feed start from implementation in the Azure Cosmos database service.
    """

    type_property_name = "Type"

    def __init__(self, start_from_type: ChangeFeedStartFromType) -> None:
        self.version = start_from_type

    @abstractmethod
    def to_dict(self) -> dict[str, Any]:
        pass

    @staticmethod
    def from_start_time(
            start_time: Optional[Union[datetime, Literal["Now", "Beginning"]]]) -> 'ChangeFeedStartFromInternal':
        if start_time is None:
            return ChangeFeedStartFromNow()
        if isinstance(start_time, datetime):
            return ChangeFeedStartFromPointInTime(start_time)
        if start_time.lower() == ChangeFeedStartFromType.NOW.value.lower():
            return ChangeFeedStartFromNow()
        if start_time.lower() == ChangeFeedStartFromType.BEGINNING.value.lower():
            return ChangeFeedStartFromBeginning()

        raise ValueError(f"Invalid start_time '{start_time}'")

    @staticmethod
    def from_json(data: dict[str, Any]) -> 'ChangeFeedStartFromInternal':
        change_feed_start_from_type = data.get(ChangeFeedStartFromInternal.type_property_name)
        if change_feed_start_from_type is None:
            raise ValueError(f"Invalid start from json [Missing {ChangeFeedStartFromInternal.type_property_name}]")

        if change_feed_start_from_type == ChangeFeedStartFromType.BEGINNING.value:
            return ChangeFeedStartFromBeginning.from_json(data)
        if change_feed_start_from_type == ChangeFeedStartFromType.LEASE.value:
            return ChangeFeedStartFromETagAndFeedRange.from_json(data)
        if change_feed_start_from_type == ChangeFeedStartFromType.NOW.value:
            return ChangeFeedStartFromNow.from_json(data)
        if change_feed_start_from_type == ChangeFeedStartFromType.POINT_IN_TIME.value:
            return ChangeFeedStartFromPointInTime.from_json(data)

        raise ValueError(f"Can not process changeFeedStartFrom for type {change_feed_start_from_type}")

    @abstractmethod
    def populate_request_headers(self, request_headers) -> None:
        pass


class ChangeFeedStartFromBeginning(ChangeFeedStartFromInternal):
    """Class for change feed start from beginning implementation in the Azure Cosmos database service.
    """

    def __init__(self) -> None:
        super().__init__(ChangeFeedStartFromType.BEGINNING)

    def to_dict(self) -> dict[str, Any]:
        return {
            self.type_property_name: ChangeFeedStartFromType.BEGINNING.value
        }

    def populate_request_headers(self, request_headers) -> None:
        pass  # there is no headers need to be set for start from beginning

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> 'ChangeFeedStartFromBeginning':
        return ChangeFeedStartFromBeginning()


class ChangeFeedStartFromETagAndFeedRange(ChangeFeedStartFromInternal):
    """Class for change feed start from etag and feed range implementation in the Azure Cosmos database service.
    """

    _etag_property_name = "Etag"
    _feed_range_property_name = "FeedRange"

    def __init__(self, etag, feed_range) -> None:
        if feed_range is None:
            raise ValueError("feed_range is missing")

        self._etag = etag
        self._feed_range = feed_range
        super().__init__(ChangeFeedStartFromType.LEASE)

    def to_dict(self) -> dict[str, Any]:
        return {
            self.type_property_name: ChangeFeedStartFromType.LEASE.value,
            self._etag_property_name: self._etag,
            self._feed_range_property_name: self._feed_range.to_dict()
        }

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> 'ChangeFeedStartFromETagAndFeedRange':
        etag = data.get(cls._etag_property_name)
        if etag is None:
            raise ValueError(f"Invalid change feed start from [Missing {cls._etag_property_name}]")

        feed_range_data = data.get(cls._feed_range_property_name)
        if feed_range_data is None:
            raise ValueError(f"Invalid change feed start from [Missing {cls._feed_range_property_name}]")
        feed_range = Range.ParseFromDict(feed_range_data)
        return cls(etag, feed_range)

    def populate_request_headers(self, request_headers) -> None:
        # change feed uses etag as the continuationToken
        if self._etag:
            request_headers[http_constants.HttpHeaders.IfNoneMatch] = self._etag


class ChangeFeedStartFromNow(ChangeFeedStartFromInternal):
    """Class for change feed start from etag and feed range implementation in the Azure Cosmos database service.
    """

    def __init__(self) -> None:
        super().__init__(ChangeFeedStartFromType.NOW)

    def to_dict(self) -> dict[str, Any]:
        return {
            self.type_property_name: ChangeFeedStartFromType.NOW.value
        }

    def populate_request_headers(self, request_headers) -> None:
        request_headers[http_constants.HttpHeaders.IfNoneMatch] = "*"

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> 'ChangeFeedStartFromNow':
        return ChangeFeedStartFromNow()


class ChangeFeedStartFromPointInTime(ChangeFeedStartFromInternal):
    """Class for change feed start from point in time implementation in the Azure Cosmos database service.
    """

    _point_in_time_ms_property_name = "PointInTimeMs"

    def __init__(self, start_time: datetime):
        if start_time is None:
            raise ValueError("start_time is missing")

        self._start_time = start_time
        super().__init__(ChangeFeedStartFromType.POINT_IN_TIME)

    def to_dict(self) -> dict[str, Any]:
        return {
            self.type_property_name: ChangeFeedStartFromType.POINT_IN_TIME.value,
            self._point_in_time_ms_property_name:
                int(self._start_time.astimezone(timezone.utc).timestamp() * 1000)
        }

    def populate_request_headers(self, request_headers) -> None:
        request_headers[http_constants.HttpHeaders.IfModified_since] =\
            self._start_time.astimezone(timezone.utc).strftime('%a, %d %b %Y %H:%M:%S GMT')

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> 'ChangeFeedStartFromPointInTime':
        point_in_time_ms = data.get(cls._point_in_time_ms_property_name)
        if point_in_time_ms is None:
            raise ValueError(f"Invalid change feed start from {cls._point_in_time_ms_property_name} ")

        point_in_time = datetime.fromtimestamp(point_in_time_ms).astimezone(timezone.utc)
        return ChangeFeedStartFromPointInTime(point_in_time)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import logging, os
from typing import Dict, Callable, Any
from azure.cosmos import CosmosClient, PartitionKey, exceptions

endpoint = os.environ["ACCOUNT_URI"]
key = os.environ["ACCOUNT_KEY"]
# Sample usage of using logging filters for diagnostics filtering
# You can filter based on request and response related attributes that are added to the log record
class CosmosStatusCodeFilter(logging.Filter):
    def filter(self, record):
        ret = (hasattr(record, 'status_code') and record.status_code > 400
               and not (record.status_code in [404, 409, 412] and getattr(record, 'sub_status_code', None) in [0, None])
               and hasattr(record, 'duration') and record.duration > 1000)
        return ret
# Initialize the logger
logger = logging.getLogger('azure.cosmos')
logger.setLevel(logging.INFO)
file_handler = logging.FileHandler('diagnostics1.output')
logger.addHandler(file_handler)
# When using the logging filter, you can set the filter directly on the logger
logger.addFilter(CosmosStatusCodeFilter())
# Initialize the Cosmos client with diagnostics enabled, no need to pass a diagnostics handler
client = CosmosClient(endpoint, key, logger=logger, enable_diagnostics_logging=True)
# Create a database and container
database_name = 'SD'
database = client.create_database_if_not_exists(id=database_name)
container_name = 'SampleContainer'
partition_key = PartitionKey(path=['/State', '/City'])
container = database.create_container_if_not_exists(id=container_name, partition_key=partition_key,
                                                    offer_throughput=400)
items = [
    {'id': '1', 'State': 'California', 'City': 'Los Angeles', 'city_level': 1},
    {'id': '2', 'State': 'Texas', 'City': 'Houston', 'city_level': 2},
    {'id': '3', 'State': 'New York', 'City': 'New York City', 'city_level': 3}
]
# Attempt to read nonexistent items to cause a 404 error
for item in items:
    try:
        container.read_item(item=str(item['id']), partition_key=[str(item['State']), str(item['City'])])
    except exceptions.CosmosHttpResponseError:
        pass

# When using the async client it can also be possible to use the logging filter with a queue logger handler
import asyncio
import queue
from queue import Queue
import logging.handlers
from azure.cosmos.aio import CosmosClient as CosmosAsyncClient
async def log_cosmos_operations():
    # Initialize the logger
    logger = logging.getLogger('azure.cosmos')
    logger.setLevel(logging.INFO)
    # Create a queue
    log_queue: Queue = queue.Queue(-1)
    # Set up the QueueHandler
    queue_handler = logging.handlers.QueueHandler(log_queue)
    # Set up the QueueListener with a FileHandler
    file_handler = logging.FileHandler('diagnostics2.output')
    file_handler.setLevel(logging.INFO)
    queue_listener = logging.handlers.QueueListener(log_queue, file_handler)
    # Configure the root logger
    logging.basicConfig(level=logging.INFO, handlers=[queue_handler])
    # Add the filter to the logger
    logger.addFilter(CosmosStatusCodeFilter())
    # Start the QueueListener
    queue_listener.start()
    # Initialize the Cosmos client with diagnostics enabled, no need to pass a diagnostics handler
    async with CosmosAsyncClient(endpoint, key, logger=logger, enable_diagnostics_logging=True) as client:
        # Create a database and container
        database_name = 'SD'
        database = await client.create_database_if_not_exists(id=database_name)
        container_name = 'SampleContainer'
        partition_key = PartitionKey(path=['/State', '/City'])
        container = await database.create_container_if_not_exists(id=container_name, partition_key=partition_key,
                                                                  offer_throughput=400)
        # Attempt to read nonexistent items to cause a 404 error
        for item in items:
            try:
                await container.read_item(item=str(item['id']), partition_key=[str(item['State']), str(item['City'])])
            except exceptions.CosmosHttpResponseError:
                pass
        # Stop the QueueListener
        queue_listener.stop()

# Run the async method
asyncio.run(log_cosmos_operations())





# The MIT License (MIT)
# Copyright (c) 2023 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Implementation of the MurmurHash3 128-bit hash function.

# MurmurHash is a non-cryptographic hash function suitable for general hash-based lookup. The name comes from two basic
# operations, multiply (MU) and rotate (R), used in its inner loop. Unlike cryptographic hash functions, it is not
# specifically designed to be difficult to reverse by an adversary, making it unsuitable for cryptographic purposes.

# This contains a Python port of the 128-bit hash function from Austin Appleby's original C++ code in SMHasher.

# This is public domain code with no copyrights. From home page of
# <a href="https://github.com/aappleby/smhasher">SMHasher</a>:
#  "All MurmurHash versions are public domain software, and the author disclaims all copyright to their code."
from ._cosmos_integers import _UInt128, _UInt64, _UInt32


def rotate_left_64(val: int, shift: int) -> int:
    return (val << shift) | (val >> (64 - shift))


def mix(value: _UInt64) -> _UInt64:
    value ^= value >> 33
    value *= 0xff51afd7ed558ccd
    value = value & 0xFFFFFFFFFFFFFFFF
    value ^= value >> 33
    value *= 0xc4ceb9fe1a85ec53
    value = value & 0xFFFFFFFFFFFFFFFF
    value ^= value >> 33
    return value


def murmurhash3_128(span: bytearray, seed: _UInt128) -> _UInt128:  # pylint: disable=too-many-statements
    """
    Python implementation of 128 bit murmurhash3 from Dot Net SDK. To match with other SDKs, It is recommended to
    do the following with number values, especially floats as other SDKs use Doubles
    -> bytearray(struct.pack("d", #)) where # represents any number. The d will treat it as a double.

    :param bytearray span:
        bytearray of value to hash
    :param _UInt128 seed:
        seed value for murmurhash3, takes in a UInt128 value from Cosmos Integers
    :return:
        The hash value as a UInt128
    :rtype:
        UInt128
    """
    c1 = _UInt64(0x87c37b91114253d5)
    c2 = _UInt64(0x4cf5ad432745937f)
    h1 = seed.get_low()
    h2 = seed.get_high()

    position = 0
    while position < len(span) - 15:
        k1 = _UInt64(int.from_bytes(span[position: position + 8], 'little'))
        k2 = _UInt64(int.from_bytes(span[position + 8: position + 16], 'little'))

        k1 *= c1
        k1.value = rotate_left_64(k1.value, 31)
        k1 *= c2
        h1 ^= k1
        h1.value = rotate_left_64(h1.value, 27)
        h1 += h2
        h1 = h1 * 5 + _UInt64(0x52dce729)

        k2 *= c2
        k2.value = rotate_left_64(k2.value, 33)
        k2 *= c1
        h2 ^= k2
        h2.value = rotate_left_64(h2.value, 31)
        h2 += h1
        h2 = h2 * 5 + _UInt64(0x38495ab5)

        position += 16

    k1 = _UInt64(0)
    k2 = _UInt64(0)
    n = len(span) & 15
    if n >= 15:
        k2 ^= _UInt64(span[position + 14] << 48)
    if n >= 14:
        k2 ^= _UInt64(span[position + 13] << 40)
    if n >= 13:
        k2 ^= _UInt64(span[position + 12] << 32)
    if n >= 12:
        k2 ^= _UInt64(span[position + 11] << 24)
    if n >= 11:
        k2 ^= _UInt64(span[position + 10] << 16)
    if n >= 10:
        k2 ^= _UInt64(span[position + 9] << 8)
    if n >= 9:
        k2 ^= _UInt64(span[position + 8] << 0)

    k2 *= c2
    k2.value = rotate_left_64(k2.value, 33)
    k2 *= c1
    h2 ^= k2

    if n >= 8:
        k1 ^= _UInt64(span[position + 7] << 56)
    if n >= 7:
        k1 ^= _UInt64(span[position + 6] << 48)
    if n >= 6:
        k1 ^= _UInt64(span[position + 5] << 40)
    if n >= 5:
        k1 ^= _UInt64(span[position + 4] << 32)
    if n >= 4:
        k1 ^= _UInt64(span[position + 3] << 24)
    if n >= 3:
        k1 ^= _UInt64(span[position + 2] << 16)
    if n >= 2:
        k1 ^= _UInt64(span[position + 1] << 8)
    if n >= 1:
        k1 ^= _UInt64(span[position + 0] << 0)

    k1 *= c1
    k1.value = rotate_left_64(k1.value, 31)
    k1 *= c2
    h1 ^= k1

    # Finalization
    h1 ^= _UInt64(len(span))
    h2 ^= _UInt64(len(span))
    h1 += h2
    h2 += h1
    h1 = mix(h1)
    h2 = mix(h2)
    h1 += h2
    h2 += h1

    return _UInt128(int(h1.value), int(h2.value))


def murmurhash3_32(data: bytearray, seed: int) -> _UInt32:
    c1: _UInt32 = _UInt32(0xcc9e2d51)
    c2: _UInt32 = _UInt32(0x1b873593)
    length: _UInt32 = _UInt32(len(data))
    h1: _UInt32 = _UInt32(seed)
    rounded_end: _UInt32 = _UInt32(length.value & 0xfffffffc)  # round down to 4 byte block

    for i in range(0, rounded_end.value, 4):
        # little endian load order
        k1: _UInt32 = _UInt32(
            (data[i] & 0xff) | ((data[i + 1] & 0xff) << 8) | ((data[i + 2] & 0xff) << 16) | (data[i + 3] << 24)
        )
        k1 *= c1
        k1.value = (k1.value << 15) | (k1.value >> 17)  # ROTL32(k1,15)
        k1 *= c2

        h1 ^= k1
        h1.value = (h1.value << 13) | (h1.value >> 19)  # ROTL32(h1,13)
        h1 = h1 * _UInt32(5) + _UInt32(0xe6546b64)

    # tail
    k1 = _UInt32(0)
    if length.value & 0x03 == 3:
        k1 ^= _UInt32((data[rounded_end.value + 2] & 0xff) << 16)
    if length.value & 0x03 >= 2:
        k1 ^= _UInt32((data[rounded_end.value + 1] & 0xff) << 8)
    if length.value & 0x03 >= 1:
        k1 ^= _UInt32(data[rounded_end.value] & 0xff)
        k1 *= c1
        k1.value = (k1.value << 15) | (k1.value >> 17)
        k1 *= c2
        h1 ^= k1

    # finalization
    h1 ^= length
    h1.value ^= h1.value >> 16
    h1 *= _UInt32(0x85ebca6b)
    h1.value ^= h1.value >> 13
    h1 *= _UInt32(0xc2b2ae35)
    h1.value ^= h1.value >> 16

    return h1



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylint: disable=missing-client-constructor-parameter-credential,missing-client-constructor-parameter-kwargs
# pylint: disable=docstring-keyword-should-match-keyword-only

"""Create, read, update and delete users in the Azure Cosmos DB SQL API service.
"""
from typing import Any, Mapping, Union, Optional, Callable

from azure.core.paging import ItemPaged
from azure.core.tracing.decorator import distributed_trace
from azure.cosmos import CosmosDict

from ._cosmos_client_connection import CosmosClientConnection
from ._base import build_options
from .permission import Permission


class UserProxy:
    """An interface to interact with a specific user.

    This class should not be instantiated directly. Instead, use the
    :func:`DatabaseProxy.get_user_client` method.

    :ivar str id:
    :ivar str user_link:
    """

    def __init__(
        self,
        client_connection: CosmosClientConnection,
        id: str,
        database_link: str,
        properties: Optional[CosmosDict] = None
    ) -> None:
        self.client_connection = client_connection
        self.id = id
        self.user_link = "{}/users/{}".format(database_link, id)
        self._properties = properties

    def __repr__(self) -> str:
        return "<UserProxy [{}]>".format(self.user_link)[:1024]

    def _get_permission_link(self, permission_or_id: Union[str, Permission, Mapping[str, Any]]) -> str:
        if isinstance(permission_or_id, str):
            return "{}/permissions/{}".format(self.user_link, permission_or_id)
        if isinstance(permission_or_id, Permission):
            return permission_or_id.permission_link
        return "{}/permissions/{}".format(self.user_link, permission_or_id["id"])

    def _get_properties(
        self
    ) -> CosmosDict:
        if self._properties is None:
            self._properties = self.read()
        return self._properties

    @distributed_trace
    def read(
        self,
        **kwargs: Any
    ) -> CosmosDict:
        """Read user properties.

        :keyword Callable response_hook: A callable invoked with the response metadata.
        :returns: A dictionary of the retrieved user properties.
        :raises ~azure.cosmos.exceptions.CosmosHttpResponseError: If the given user couldn't be retrieved.
        :rtype: ~azure.cosmos.CosmosDict[str, Any]
        """
        request_options = build_options(kwargs)
        self._properties = self.client_connection.ReadUser(
            user_link=self.user_link,
            options=request_options,
            **kwargs
        )
        return self._properties

    @distributed_trace
    def list_permissions(
            self,
            max_item_count: Optional[int] = None,
            *,
            response_hook: Optional[Callable[[Mapping[str, Any], ItemPaged[dict[str, Any]]], None]] = None,
            **kwargs: Any
    ) -> ItemPaged[dict[str, Any]]:
        """List all permission for the user.

        :param int max_item_count: Max number of permissions to be returned in the enumeration operation.
        :keyword response_hook: A callable invoked with the response metadata.
        :paramtype response_hook: Callable[[Mapping[str, Any], ItemPaged[dict[str, Any]]], None]
        :returns: An Iterable of permissions (dicts).
        :rtype: Iterable[dict[str, Any]]
        """
        feed_options = build_options(kwargs)
        if max_item_count is not None:
            feed_options["maxItemCount"] = max_item_count

        result = self.client_connection.ReadPermissions(
            user_link=self.user_link,
            options=feed_options,
            response_hook=response_hook,
            **kwargs)

        if response_hook:
            response_hook(self.client_connection.last_response_headers, result)

        return result

    @distributed_trace
    def query_permissions(
        self,
        query: str,
        parameters: Optional[list[dict[str, Any]]] = None,
        max_item_count: Optional[int] = None,
        *,
        response_hook: Optional[Callable[[Mapping[str, Any], ItemPaged[dict[str, Any]]], None]] = None,
        **kwargs: Any
    ) -> ItemPaged[dict[str, Any]]:
        """Return all permissions matching the given `query`.

        :param str query: The Azure Cosmos DB SQL query to execute.
        :param parameters: Optional array of parameters to the query. Ignored if no query is provided.
        :type parameters: list[dict[str, Any]]
        :param int max_item_count: Max number of permissions to be returned in the enumeration operation.
        :keyword response_hook: A callable invoked with the response metadata.
        :paramtype response_hook: Callable[[Mapping[str, Any], ItemPaged[dict[str, Any]]], None]
        :returns: An Iterable of permissions (dicts).
        :rtype: Iterable[dict[str, Any]]
        """
        feed_options = build_options(kwargs)
        if max_item_count is not None:
            feed_options["maxItemCount"] = max_item_count

        result = self.client_connection.QueryPermissions(
            user_link=self.user_link,
            query=query if parameters is None else {"query": query, "parameters": parameters},
            options=feed_options,
            response_hook=response_hook,
            **kwargs
        )

        if response_hook:
            response_hook(self.client_connection.last_response_headers, result)

        return result

    @distributed_trace
    def get_permission(
        self,
        permission: Union[str, Permission, Mapping[str, Any]],
        **kwargs: Any
    ) -> Permission:
        """Get the permission identified by `id`.

        :param permission: The ID (name), dict representing the properties or :class:`~azure.cosmos.Permission`
            instance of the permission to be retrieved.
        :type permission: Union[str, ~azure.cosmos.Permission, dict[str, Any]]
        :keyword Callable response_hook: A callable invoked with the response metadata.
        :returns: A dict representing the retrieved permission.
        :raises ~azure.cosmos.exceptions.CosmosHttpResponseError: If the given permission couldn't be retrieved.
        :rtype: dict[str, Any]
        """
        request_options = build_options(kwargs)
        permission_resp = self.client_connection.ReadPermission(
            permission_link=self._get_permission_link(permission),
            options=request_options,
            **kwargs
        )
        return Permission(
            id=permission_resp["id"],
            user_link=self.user_link,
            permission_mode=permission_resp["permissionMode"],
            resource_link=permission_resp["resource"],
            properties=permission_resp,
        )

    @distributed_trace
    def create_permission(self, body: dict[str, Any], **kwargs: Any) -> Permission:
        """Create a permission for the user.

        To update or replace an existing permision, use the :func:`UserProxy.upsert_permission` method.

        :param dict[str, Any] body: A dict-like object representing the permission to create.
        :keyword Callable response_hook: A callable invoked with the response metadata.
        :returns: A dict representing the new permission.
        :raises ~azure.cosmos.exceptions.CosmosHttpResponseError: If the given permission couldn't be created.
        :rtype: dict[str, Any]
        """
        request_options = build_options(kwargs)
        permission = self.client_connection.CreatePermission(
            user_link=self.user_link,
            permission=body,
            options=request_options,
            **kwargs
        )
        return Permission(
            id=permission["id"],
            user_link=self.user_link,
            permission_mode=permission["permissionMode"],
            resource_link=permission["resource"],
            properties=permission,
        )

    @distributed_trace
    def upsert_permission(self, body: dict[str, Any], **kwargs: Any) -> Permission:
        """Insert or update the specified permission.

        If the permission already exists in the container, it is replaced. If
        the permission does not exist, it is inserted.

        :param dict[str, Any] body: A dict-like object representing the permission to update or insert.
        :keyword Callable response_hook: A callable invoked with the response metadata.
        :returns: A dict representing the upserted permission.
        :raises ~azure.cosmos.exceptions.CosmosHttpResponseError: If the given permission could not be upserted.
        :rtype: dict[str, Any]
        """
        request_options = build_options(kwargs)
        permission = self.client_connection.UpsertPermission(
            user_link=self.user_link, permission=body, options=request_options, **kwargs
        )
        return Permission(
            id=permission["id"],
            user_link=self.user_link,
            permission_mode=permission["permissionMode"],
            resource_link=permission["resource"],
            properties=permission,
        )

    @distributed_trace
    def replace_permission(
        self,
        permission: Union[str, Permission, Mapping[str, Any]],
        body: dict[str, Any],
        **kwargs
    ) -> Permission:
        """Replaces the specified permission if it exists for the user.

        If the permission does not already exist, an exception is raised.

        :param permission: The ID (name), dict representing the properties or :class:`~azure.cosmos.Permission`
            instance of the permission to be replaced.
        :type permission: Union[str, ~azure.cosmos.Permission, dict[str, Any]]
        :param dict[str, Any] body: A dict-like object representing the permission to replace.
        :keyword Callable response_hook: A callable invoked with the response metadata.
        :returns: A dict representing the permission after replace went through.
        :raises ~azure.cosmos.exceptions.CosmosHttpResponseError: If the replace operation failed or the permission
            with given id does not exist.
        :rtype: dict[str, Any]
        """
        request_options = build_options(kwargs)
        permission_resp = self.client_connection.ReplacePermission(
            permission_link=self._get_permission_link(permission),
            permission=body,
            options=request_options,
            **kwargs
        )
        return Permission(
            id=permission_resp["id"],
            user_link=self.user_link,
            permission_mode=permission_resp["permissionMode"],
            resource_link=permission_resp["resource"],
            properties=permission_resp,
        )

    @distributed_trace
    def delete_permission(
        self,
        permission: Union[str, Permission, Mapping[str, Any]],
        **kwargs
    ) -> None:
        """Delete the specified permission from the user.

        If the permission does not already exist, an exception is raised.

        :param permission: The ID (name), dict representing the properties or :class:`~azure.cosmos.Permission`
            instance of the permission to be replaced.
        :type permission: Union[str, ~azure.cosmos.Permission, dict[str, Any]]
        :keyword Callable response_hook: A callable invoked with the response metadata.
        :raises ~azure.cosmos.exceptions.CosmosHttpResponseError: The permission wasn't deleted successfully.
        :raises ~azure.cosmos.exceptions.CosmosResourceNotFoundError: The permission does not exist for the user.
        :rtype: None
        """
        request_options = build_options(kwargs)
        self.client_connection.DeletePermission(
            permission_link=self._get_permission_link(permission), options=request_options, **kwargs
        )



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import base64
import json
import time
import os
import unittest
from io import StringIO

import pytest
from azure.core.credentials import AccessToken

import test_config
from azure.cosmos import exceptions
from azure.cosmos.aio import CosmosClient, DatabaseProxy, ContainerProxy
from azure.core.exceptions import HttpResponseError

def _remove_padding(encoded_string):
    while encoded_string.endswith("="):
        encoded_string = encoded_string[0:len(encoded_string) - 1]

    return encoded_string


def get_test_item(num):
    test_item = {
        'pk': 'pk',
        'id': 'Item_' + str(num),
        'test_object': True,
        'lastName': 'Smith'
    }
    return test_item


class CosmosEmulatorCredential(object):
    async def get_token(self, *scopes, **kwargs):
        # type: (*str, **Any) -> AccessToken
        """Request an access token for the emulator. Based on Azure Core's Access Token Credential.

        This method is called automatically by Azure SDK clients.

        :param str scopes: desired scopes for the access token. This method requires at least one scope.
        :rtype: :class:`azure.core.credentials.AccessToken`
        :raises CredentialUnavailableError: the credential is unable to attempt authentication because it lacks
          required data, state, or platform support
        :raises ~azure.core.exceptions.ClientAuthenticationError: authentication failed. The error's ``message``
          attribute gives a reason.
        """
        aad_header_cosmos_emulator = "{\"typ\":\"JWT\",\"alg\":\"RS256\",\"x5t\":\"" \
                                     "CosmosEmulatorPrimaryMaster\",\"kid\":\"CosmosEmulatorPrimaryMaster\"}"

        aad_claim_cosmos_emulator_format = {"aud": "https://localhost.localhost",
                                            "iss": "https://sts.fake-issuer.net/7b1999a1-dfd7-440e-8204-00170979b984",
                                            "iat": int(time.time()), "nbf": int(time.time()),
                                            "exp": int(time.time() + 7200), "aio": "", "appid": "localhost",
                                            "appidacr": "1", "idp": "https://localhost:8081/",
                                            "oid": "96313034-4739-43cb-93cd-74193adbe5b6", "rh": "", "sub": "localhost",
                                            "tid": "EmulatorFederation", "uti": "", "ver": "1.0",
                                            "scp": "user_impersonation",
                                            "groups": ["7ce1d003-4cb3-4879-b7c5-74062a35c66e",
                                                       "e99ff30c-c229-4c67-ab29-30a6aebc3e58",
                                                       "5549bb62-c77b-4305-bda9-9ec66b85d9e4",
                                                       "c44fd685-5c58-452c-aaf7-13ce75184f65",
                                                       "be895215-eab5-43b7-9536-9ef8fe130330"]}

        emulator_key = test_config.TestConfig.masterKey

        first_encoded_bytes = base64.urlsafe_b64encode(aad_header_cosmos_emulator.encode("utf-8"))
        first_encoded_padded = str(first_encoded_bytes, "utf-8")
        first_encoded = _remove_padding(first_encoded_padded)

        str_io_obj = StringIO()
        json.dump(aad_claim_cosmos_emulator_format, str_io_obj)
        aad_claim_cosmos_emulator_format_string = str(str_io_obj.getvalue()).replace(" ", "")
        second = aad_claim_cosmos_emulator_format_string
        second_encoded_bytes = base64.urlsafe_b64encode(second.encode("utf-8"))
        second_encoded_padded = str(second_encoded_bytes, "utf-8")
        second_encoded = _remove_padding(second_encoded_padded)

        emulator_key_encoded_bytes = base64.urlsafe_b64encode(emulator_key.encode("utf-8"))
        emulator_key_encoded_padded = str(emulator_key_encoded_bytes, "utf-8")
        emulator_key_encoded = _remove_padding(emulator_key_encoded_padded)

        return AccessToken(first_encoded + "." + second_encoded + "." + emulator_key_encoded, int(time.time() + 7200))


@pytest.mark.cosmosEmulator
class TestAADAsync(unittest.IsolatedAsyncioTestCase):
    client: CosmosClient = None
    database: DatabaseProxy = None
    container: ContainerProxy = None
    configs = test_config.TestConfig
    host = configs.host
    masterKey = configs.masterKey
    credential = CosmosEmulatorCredential() if configs.is_emulator else configs.credential_async

    @classmethod
    def setUpClass(cls):
        if (cls.credential == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")

    async def asyncSetUp(self):
        self.client = CosmosClient(self.host, self.credential)
        self.database = self.client.get_database_client(self.configs.TEST_DATABASE_ID)
        self.container = self.database.get_container_client(self.configs.TEST_SINGLE_PARTITION_CONTAINER_ID)

    async def asyncTearDown(self):
        await self.client.close()

    async def test_aad_credentials_async(self):
        # Do any R/W data operations with your authorized AAD client

        print("Container info: " + str(await self.container.read()))
        await self.container.create_item(get_test_item(0))
        print("Point read result: " + str(await self.container.read_item(item='Item_0', partition_key='pk')))
        query_results = [item async for item in self.container.query_items(query='select * from c', partition_key='pk')]
        assert len(query_results) == 1
        print("Query result: " + str(query_results[0]))
        await self.container.delete_item(item='Item_0', partition_key='pk')

        # Attempting to do management operations will return a 403 Forbidden exception
        try:
            await self.client.delete_database(self.configs.TEST_DATABASE_ID)
        except exceptions.CosmosHttpResponseError as e:
            assert e.status_code == 403
            print("403 error assertion success")

    async def _run_with_scope_capture_async(self, credential_cls, action):
        scopes_captured = []

        orig_get_token = credential_cls.get_token

        async def capturing_get_token(self, *scopes, **kwargs):
            scopes_captured.extend(scopes)
            return await orig_get_token(self, *scopes, **kwargs)

        credential_cls.get_token = capturing_get_token
        try:
            result = await action(scopes_captured)
            return scopes_captured, result
        finally:
            credential_cls.get_token = orig_get_token

    async def test_override_scope_no_fallback_async(self):
        """When override scope is provided, only that scope is used and no fallback occurs."""
        override_scope = "https://my.custom.scope/.default"
        os.environ["AZURE_COSMOS_AAD_SCOPE_OVERRIDE"] = override_scope

        async def action(scopes_captured):
            credential = CosmosEmulatorCredential()
            client = CosmosClient(self.host, credential)
            try:
                db = client.get_database_client(self.configs.TEST_DATABASE_ID)
                container = db.get_container_client(self.configs.TEST_SINGLE_PARTITION_CONTAINER_ID)
                await container.create_item(get_test_item(20))
                return container
            finally:
                await client.close()

        scopes, container = await self._run_with_scope_capture_async(CosmosEmulatorCredential, action)
        try:
            assert all(scope == override_scope for scope in scopes), f"Expected only override scope, got: {scopes}"
        finally:
            del os.environ["AZURE_COSMOS_AAD_SCOPE_OVERRIDE"]
            try:
                await container.delete_item(item='Item_20', partition_key='pk')
            except Exception:
                pass

    async def test_override_scope_no_fallback_on_error_async(self):
        """When override scope is provided and auth fails, no fallback occurs."""
        override_scope = "https://my.custom.scope/.default"
        os.environ["AZURE_COSMOS_AAD_SCOPE_OVERRIDE"] = override_scope

        class FailingCredential(CosmosEmulatorCredential):
            async def get_token(self, *scopes, **kwargs):
                raise Exception("AADSTS500011: Simulated error for override scope")

        async def action(scopes_captured):
            credential = FailingCredential()
            client = CosmosClient(self.host, credential)
            try:
                db = client.get_database_client(self.configs.TEST_DATABASE_ID)
                container = db.get_container_client(self.configs.TEST_SINGLE_PARTITION_CONTAINER_ID)
                try:
                    await container.create_item(get_test_item(21))
                except Exception:
                    pass
                return container
            finally:
                await client.close()

        scopes, container = await self._run_with_scope_capture_async(FailingCredential, action)
        try:
            assert all(scope == override_scope for scope in scopes), f"Expected only override scope, got: {scopes}"
        finally:
            del os.environ["AZURE_COSMOS_AAD_SCOPE_OVERRIDE"]
            try:
                await container.delete_item(item='Item_21', partition_key='pk')
            except Exception:
                pass

    async def test_account_scope_only_async(self):
        """When account scope is provided, only that scope is used."""
        account_scope = "https://localhost/.default"
        os.environ["AZURE_COSMOS_AAD_SCOPE_OVERRIDE"] = ""

        async def action(scopes_captured):
            credential = CosmosEmulatorCredential()
            client = CosmosClient(self.host, credential)
            try:
                db = client.get_database_client(self.configs.TEST_DATABASE_ID)
                container = db.get_container_client(self.configs.TEST_SINGLE_PARTITION_CONTAINER_ID)
                await container.create_item(get_test_item(22))
                return container
            finally:
                await client.close()

        scopes, container = await self._run_with_scope_capture_async(CosmosEmulatorCredential, action)
        try:
            assert all(scope == account_scope for scope in scopes), f"Expected only account scope, got: {scopes}"
        finally:
            try:
                await container.delete_item(item='Item_22', partition_key='pk')
            except Exception:
                pass

    async def test_account_scope_fallback_on_error_async(self):
        """When account scope is provided and auth fails, fallback to default scope occurs."""
        account_scope = "https://localhost/.default"
        fallback_scope = "https://cosmos.azure.com/.default"
        os.environ["AZURE_COSMOS_AAD_SCOPE_OVERRIDE"] = ""

        class FallbackCredential(CosmosEmulatorCredential):
            def __init__(self):
                self.call_count = 0

            async def get_token(self, *scopes, **kwargs):
                self.call_count += 1
                if self.call_count == 1:
                    raise HttpResponseError(message="AADSTS500011: Simulated error for fallback")
                return await super().get_token(*scopes, **kwargs)

        async def action(scopes_captured):
            credential = FallbackCredential()
            client = CosmosClient(self.host, credential)
            try:
                db = client.get_database_client(self.configs.TEST_DATABASE_ID)
                container = db.get_container_client(self.configs.TEST_SINGLE_PARTITION_CONTAINER_ID)
                await container.create_item(get_test_item(23))
                return container
            finally:
                await client.close()

        scopes, container = await self._run_with_scope_capture_async(FallbackCredential, action)
        try:
            assert account_scope in scopes and fallback_scope in scopes, f"Expected fallback to default scope, got: {scopes}"
        finally:
            try:
                await container.delete_item(item='Item_23', partition_key='pk')
            except Exception:
                pass

if __name__ == "__main__":
    unittest.main()



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Session Consistency Tracking in the Azure Cosmos database service.
"""

import logging
import sys
import traceback
import threading
from typing import Any, Optional

from . import _base
from . import http_constants
from ._routing.routing_map_provider import SmartRoutingMapProvider
from ._routing.aio.routing_map_provider import SmartRoutingMapProvider as SmartRoutingMapProviderAsync
from ._vector_session_token import VectorSessionToken
from .exceptions import CosmosHttpResponseError
from .partition_key import PartitionKey

# pylint: disable=protected-access,too-many-nested-blocks

logger = logging.getLogger("azure.cosmos.SessionContainer")

class SessionContainer(object):
    def __init__(self):
        self.collection_name_to_rid = {}
        self.rid_to_session_token = {}
        self.session_lock = threading.RLock()

    def get_session_token(
            self,
            resource_path: str,
            pk_value: Any,
            container_properties_cache: dict[str, dict[str, Any]],
            routing_map_provider: SmartRoutingMapProvider,
            partition_key_range_id: Optional[int],
            options: dict[str, Any]
    ) -> str:
        """Get Session Token for the given collection and partition key information.

        :param str resource_path: Self link / path to the resource
        :param ~azure.cosmos.SmartRoutingMapProvider routing_map_provider: routing map containing relevant session
            information, such as partition key ranges for a given collection
        :param Any pk_value: The partition key value being used for the operation
        :param container_properties_cache: Container properties cache used to fetch partition key definitions
        :type container_properties_cache: dict[str, dict[str, Any]]
        :param int partition_key_range_id: The partition key range ID used for the operation
        :param options: Options for the operation calling this method
        :type options: dict[str, Any]
        :return: Session Token dictionary for the collection_id, will be empty string if not found or if the operation
        does not require a session token (single master write operations).
        :rtype: str
        """

        with self.session_lock:
            is_name_based = _base.IsNameBased(resource_path)
            session_token = ""
            collection_name = ""

            try:
                if is_name_based:
                    # get the collection name
                    collection_name = _base.GetItemContainerLink(resource_path)
                    if collection_name in self.collection_name_to_rid:
                        # if the collection name is already in the map, use the rid from there
                        collection_rid = self.collection_name_to_rid[collection_name]
                    else:
                        # if the collection name is not in the map, we need to get the rid from containers cache
                        collection_rid = container_properties_cache.get(collection_name, {}).get("_rid")
                        if collection_rid:
                            self.collection_name_to_rid[collection_name] = collection_rid
                else:
                    collection_rid = _base.GetItemContainerLink(resource_path)

                if collection_rid in self.rid_to_session_token and (collection_name in container_properties_cache or
                                                                    collection_rid in container_properties_cache):
                    token_dict = self.rid_to_session_token[collection_rid]
                    if partition_key_range_id is not None:
                        # if we find a cached session token for the relevant pk range id, use that session token
                        if token_dict.get(partition_key_range_id):
                            vector_session_token = token_dict.get(partition_key_range_id)
                            session_token = "{0}:{1}".format(partition_key_range_id, vector_session_token.session_token)
                        # if we don't find it, we do a session token merge for the parent pk ranges
                        # this should only happen immediately after a partition split
                        else:
                            container_routing_map = \
                                routing_map_provider._collection_routing_map_by_item[collection_name]
                            current_range = container_routing_map._rangeById.get(partition_key_range_id)
                            if current_range is not None:
                                vector_session_token = self._resolve_partition_local_session_token(current_range,
                                                                                                   token_dict)
                                if vector_session_token is not None:
                                    session_token = "{0}:{1}".format(partition_key_range_id, vector_session_token)
                    elif pk_value is not None:
                        collection_pk_definition = container_properties_cache[collection_name]["partitionKey"]
                        partition_key = PartitionKey(path=collection_pk_definition['paths'],
                                                     kind=collection_pk_definition['kind'],
                                                     version=collection_pk_definition['version'])
                        epk_range = partition_key._get_epk_range_for_partition_key(pk_value=pk_value)
                        pk_range = routing_map_provider.get_overlapping_ranges(collection_name,
                                                                               [epk_range],
                                                                               options)
                        if len(pk_range) > 0:
                            partition_key_range_id = pk_range[0]['id']
                            vector_session_token = self._resolve_partition_local_session_token(pk_range, token_dict)
                            if vector_session_token is not None:
                                session_token = "{0}:{1}".format(partition_key_range_id, vector_session_token)
                    else:
                        # we're executing a cross partition streamable query that can be resolved by the gateway
                        # send the entire compound session token for the container to target all partitions
                        # TODO: this logic breaks large containers, needs to be addressed along with requesting
                        #  a query plan for every query
                        session_token_list = []
                        for key in token_dict.keys():
                            session_token_list.append("{0}:{1}".format(key, token_dict[key].convert_to_string()))
                        session_token = ",".join(session_token_list)
                    return session_token
                return ""
            except (KeyError, AttributeError) as e:  # pylint: disable=broad-except
                logger.debug("Error while resolving session token: %s", e)
                return ""

    async def get_session_token_async(
            self,
            resource_path: str,
            pk_value: Any,
            container_properties_cache: dict[str, dict[str, Any]],
            routing_map_provider: SmartRoutingMapProviderAsync,
            partition_key_range_id: Optional[str],
            options: dict[str, Any]
    ) -> str:
        """Get Session Token for the given collection and partition key information.

        :param str resource_path: Self link / path to the resource
        :param ~azure.cosmos.SmartRoutingMapProviderAsync routing_map_provider: routing map containing relevant session
            information, such as partition key ranges for a given collection
        :param Any pk_value: The partition key value being used for the operation
        :param container_properties_cache: Container properties cache used to fetch partition key definitions
        :type container_properties_cache: dict[str, dict[str, Any]]
        :param Any routing_map_provider: The routing map provider containing the partition key range cache logic
        :param str partition_key_range_id: The partition key range ID used for the operation
        :param options: Options for the operation calling this method
        :type options: dict[str, Any]
        :return: Session Token dictionary for the collection_id, will be empty string if not found or if the operation
        does not require a session token (single master write operations).
        :rtype: str
        """

        with self.session_lock:
            is_name_based = _base.IsNameBased(resource_path)
            session_token = ""

            try:
                if is_name_based:
                    # get the collection name
                    collection_name = _base.GetItemContainerLink(resource_path)
                    if collection_name in self.collection_name_to_rid:
                        # if the collection name is already in the map, use the rid from there
                        collection_rid = self.collection_name_to_rid[collection_name]
                    else:
                        # if the collection name is not in the map, we need to get the rid from containers cache
                        collection_rid = container_properties_cache.get(collection_name, {}).get("_rid")
                        if collection_rid:
                            self.collection_name_to_rid[collection_name] = collection_rid
                else:
                    collection_rid = _base.GetItemContainerLink(resource_path)

                if collection_rid in self.rid_to_session_token and (collection_name in container_properties_cache or
                                                                    collection_rid in container_properties_cache):
                    token_dict = self.rid_to_session_token[collection_rid]
                    if partition_key_range_id is not None:
                        # if we find a cached session token for the relevant pk range id, use that session token
                        if token_dict.get(partition_key_range_id):
                            vector_session_token = token_dict.get(partition_key_range_id)
                            session_token = "{0}:{1}".format(partition_key_range_id,
                                                             vector_session_token.session_token)
                        # if we don't find it, we do a session token merge for the parent pk ranges
                        # this should only happen immediately after a partition split
                        else:
                            container_routing_map = \
                                routing_map_provider._collection_routing_map_by_item[collection_name]
                            current_range = container_routing_map._rangeById.get(partition_key_range_id)
                            if current_range is not None:
                                vector_session_token = self._resolve_partition_local_session_token(current_range,
                                                                                                   token_dict)
                                if vector_session_token is not None:
                                    session_token = "{0}:{1}".format(partition_key_range_id, vector_session_token)
                    elif pk_value is not None:
                        collection_pk_definition = container_properties_cache[collection_name]["partitionKey"]
                        partition_key = PartitionKey(path=collection_pk_definition['paths'],
                                                     kind=collection_pk_definition['kind'],
                                                     version=collection_pk_definition['version'])
                        epk_range = partition_key._get_epk_range_for_partition_key(pk_value=pk_value)
                        pk_range = await routing_map_provider.get_overlapping_ranges(collection_name,
                                                                                     [epk_range],
                                                                                     options)
                        if len(pk_range) > 0:
                            partition_key_range_id = pk_range[0]['id']
                            vector_session_token = self._resolve_partition_local_session_token(pk_range, token_dict)
                            if vector_session_token is not None:
                                session_token = "{0}:{1}".format(partition_key_range_id, vector_session_token)
                    else:
                        # we're executing a cross partition streamable query that can be resolved by the gateway
                        # send the entire compound session token for the container to target all partitions
                        # TODO: this logic breaks large containers, needs to be addressed along with requesting
                        #  a query plan for every query
                        session_token_list = []
                        for key in token_dict.keys():
                            session_token_list.append("{0}:{1}".format(key, token_dict[key].convert_to_string()))
                        session_token = ",".join(session_token_list)
                    return session_token
                return ""
            except Exception:  # pylint: disable=broad-except
                return ""

    def set_session_token(self, client_connection, response_result, response_headers):
        """Session token must only be updated from response of requests that
        successfully mutate resource on the server side (write, replace, delete etc).

        :param client_connection: Client connection used to refresh the partition key range cache if needed
        :type client_connection: Union[azure.cosmos.CosmosClientConnection, azure.cosmos.aio.CosmosClientConnection]
        :param dict response_result:
        :param dict response_headers:
        :return: None
        """
        # pylint: disable=too-many-statements

        # there are two pieces of information that we need to update session token-
        # self link which has the rid representation of the resource, and
        # x-ms-alt-content-path which is the string representation of the resource

        with self.session_lock:
            try:
                self_link = response_result.get("_self")

                # extract alternate content path from the response_headers
                # (only document level resource updates will have this),
                # and if not present, then we can assume that we don't have to update
                # session token for this request
                alt_content_path_key = http_constants.HttpHeaders.AlternateContentPath
                response_result_id_key = "id"
                response_result_id = None
                if alt_content_path_key in response_headers:
                    alt_content_path = response_headers[http_constants.HttpHeaders.AlternateContentPath]
                    if response_result_id_key in response_result:
                        response_result_id = response_result[response_result_id_key]
                else:
                    return
                if self_link is not None:
                    collection_rid, collection_name = _base.GetItemContainerInfo(self_link, alt_content_path,
                                                                                response_result_id)
                else:
                    # if for whatever reason we don't have a _self link at this point, we use the container name
                    collection_name = alt_content_path
                    collection_rid = self.collection_name_to_rid.get(collection_name)
                # if the response came in with a new partition key range id after a split, refresh the pk range cache
                partition_key_range_id = response_headers.get(http_constants.HttpHeaders.PartitionKeyRangeID)
                collection_ranges = None
                if client_connection:
                    collection_ranges = \
                        client_connection._routing_map_provider._collection_routing_map_by_item.get(collection_name)
                if collection_ranges and not collection_ranges._rangeById.get(partition_key_range_id):
                    client_connection.refresh_routing_map_provider()
            except ValueError:
                return
            except Exception:  # pylint: disable=broad-except
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback, limit=2, file=sys.stdout)
                return

            if collection_name in self.collection_name_to_rid:
                # check if the rid for the collection name has changed
                # this means that potentially, the collection was deleted
                # and recreated
                existing_rid = self.collection_name_to_rid[collection_name]
                if collection_rid != existing_rid:
                    # flush the session tokens for the old rid, and
                    # update the new rid into the collection name to rid map.
                    self.rid_to_session_token[existing_rid] = {}
                    self.collection_name_to_rid[collection_name] = collection_rid

            # parse session token
            parsed_tokens = self.parse_session_token(response_headers)

            # update session token in collection rid to session token map
            if collection_rid in self.rid_to_session_token:
                # we need to update the session tokens for 'this' collection
                for id_ in parsed_tokens:  # pylint: disable=consider-using-dict-items
                    old_session_token = (
                        self.rid_to_session_token[collection_rid][id_]
                        if id_ in self.rid_to_session_token[collection_rid]
                        else None
                    )
                    if not old_session_token:
                        self.rid_to_session_token[collection_rid][id_] = parsed_tokens[id_]
                    else:
                        self.rid_to_session_token[collection_rid][id_] = parsed_tokens[id_].merge(old_session_token)
            else:
                self.rid_to_session_token[collection_rid] = parsed_tokens
            self.collection_name_to_rid[collection_name] = collection_rid

    def clear_session_token(self, response_headers):
        with self.session_lock:
            collection_rid = ""
            alt_content_path = ""
            alt_content_path_key = http_constants.HttpHeaders.AlternateContentPath
            if alt_content_path_key in response_headers:
                alt_content_path = response_headers[http_constants.HttpHeaders.AlternateContentPath]
                if alt_content_path in self.collection_name_to_rid:
                    collection_rid = self.collection_name_to_rid[alt_content_path]
                    del self.collection_name_to_rid[alt_content_path]
                    del self.rid_to_session_token[collection_rid]

    @staticmethod
    def parse_session_token(response_headers):
        """Extracts session token from response headers and parses.

        :param dict response_headers:
        :return: A dictionary of partition id to session lsn for given collection
        :rtype: dict
        """

        # extract session token from response header
        session_token = ""
        if http_constants.HttpHeaders.SessionToken in response_headers:
            session_token = response_headers[http_constants.HttpHeaders.SessionToken]

        id_to_sessionlsn = {}
        if session_token:
            # extract id, lsn from the token. For p-collection,
            # the token will be a concatenation of pairs for each collection
            token_pairs = session_token.split(",")
            for token_pair in token_pairs:
                tokens = token_pair.split(":")
                if len(tokens) == 2:
                    id_ = tokens[0]
                    sessionToken = VectorSessionToken.create(tokens[1])
                    if sessionToken is None:
                        raise CosmosHttpResponseError(
                            status_code=http_constants.StatusCodes.INTERNAL_SERVER_ERROR,
                            message="Could not parse the received session token: %s" % tokens[1],
                        )
                    id_to_sessionlsn[id_] = sessionToken
        return id_to_sessionlsn

    def _resolve_partition_local_session_token(self, pk_range, token_dict):
        parent_session_token = None
        parents = pk_range[0].get('parents').copy()
        parents.append(pk_range[0]['id'])
        for parent in parents:
            session_token = token_dict.get(parent)
            if session_token is not None:
                vector_session_token = session_token.session_token
                if parent_session_token is None:
                    parent_session_token = vector_session_token
                # if initial token is already set, and the next parent's token is cached, merge vector session tokens
                else:
                    vector_token_1 = VectorSessionToken.create(parent_session_token)
                    vector_token_2 = VectorSessionToken.create(vector_session_token)
                    vector_token = vector_token_1.merge(vector_token_2)
                    parent_session_token = vector_token.session_token
        return parent_session_token


class Session(object):
    """State of an Azure Cosmos session.

    This session object can be shared across clients within the same process.

    :param url_connection:
    """

    def __init__(self, url_connection):
        self.url_connection = url_connection
        self.session_container = SessionContainer()
        # include creation time, and some other stats

    def clear_session_token(self, response_headers):
        self.session_container.clear_session_token(response_headers)

    def update_session(self, client_connection, response_result, response_headers):
        self.session_container.set_session_token(client_connection, response_result, response_headers)

    def get_session_token(self, resource_path, pk_value, container_properties_cache, routing_map_provider,
                          partition_key_range_id, options):
        return self.session_container.get_session_token(resource_path, pk_value, container_properties_cache,
                                                        routing_map_provider, partition_key_range_id, options)

    async def get_session_token_async(self, resource_path, pk_value, container_properties_cache, routing_map_provider,
                                      partition_key_range_id, options):
        return await self.session_container.get_session_token_async(resource_path,
                                                                    pk_value,
                                                                    container_properties_cache,
                                                                    routing_map_provider,
                                                                    partition_key_range_id,
                                                                    options)



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import unittest

import pytest

import azure.cosmos.diagnostics as m

_common = {
    'x-ms-activity-id',
    'x-ms-session-token',
    'x-ms-item-count',
    'x-ms-request-quota',
    'x-ms-resource-usage',
    'x-ms-retry-after-ms',
}

_headers = dict(zip(_common, _common))
_headers['other'] = 'other'


@pytest.mark.cosmosEmulator
class TestOldDiagnostics(unittest.TestCase):

    def test_init(self):
        rh = m.RecordDiagnostics()
        assert rh.headers == {}

    def test_headers(self):
        rh = m.RecordDiagnostics()
        rh(_headers, "body")
        assert rh.headers == _headers
        assert rh.headers is not _headers

    def test_headers_case(self):
        rh = m.RecordDiagnostics()
        rh(_headers, "body")
        rh_headers = rh.headers
        for key in rh.headers.keys():
            assert key.upper() in rh_headers
            assert key.lower() in rh_headers

    def test_common_attrs(self):
        rh = m.RecordDiagnostics()
        rh(_headers, "body")
        for name in _common:
            assert rh.headers[name] == name
            attr = name.replace('x-ms-', '').replace('-', '_')
            assert getattr(rh, attr) == name

    def test_other_attrs(self):
        rh = m.RecordDiagnostics()
        rh(_headers, "body")
        assert rh.headers['other'] == 'other'
        with self.assertRaises(AttributeError):
            rh.other


if __name__ == '__main__':
    unittest.main()
